"""

"""
# Partial imports
from abc import ABC, abstractmethod
from jsonschema import Draft7Validator
from timeit import default_timer as timer
from typing import Type, Dict, List, Tuple, Union

# Imports from internal modules
from .instance import InstanceCore
from .solution import SolutionCore
from .experiment import ExperimentCore

from cornflow_client.constants import (
    STATUS_OPTIMAL,
    STATUS_NOT_SOLVED,
    STATUS_INFEASIBLE,
    STATUS_UNDEFINED,
    STATUS_TIME_LIMIT,
    SOLUTION_STATUS_FEASIBLE,
    SOLUTION_STATUS_INFEASIBLE,
    NoSolverException,
    BadConfiguration,
    BadSolution,
    BadInstance,
)


class ApplicationCore(ABC):
    """
    The application template.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Mandatory property

        :return: the name of the class.
        """
        raise NotImplementedError()

    @property
    def description(self) -> Union[str, None]:
        """
        Optional property

        :return: the description of the class
        """
        return None

    @property
    @abstractmethod
    def instance(self) -> Type[InstanceCore]:
        """
        Mandatory property

        :return: the constructor for the instance.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def solution(self) -> Type[SolutionCore]:
        """
        Mandatory property

        :return: the constructor for the solution.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def schema(self) -> dict:
        """
        Mandatory property

        :return: the configuration schema used for the solve() method.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def test_cases(self) -> List[Union[Dict, Tuple[Dict, Dict]]]:
        """
        Mandatory property

        :return: a list of datasets following the json-schema.
          if each element in the list is:

          * **dict**: each element is an instance
          * **tuple**: the first part is the instance, the second its solution
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def solvers(self) -> Dict[str, Type[ExperimentCore]]:
        """
        Mandatory property

        :return: a dictionary of constructors for solution methods for this particular problem.
        """
        raise NotImplementedError()

    def solve(
        self, data: dict, config: dict, solution_data: dict = None
    ) -> Tuple[Dict, Union[Dict, None], Union[Dict, None], str, Dict]:
        """
        :param data: json for the problem
        :param config: execution configuration, including solver
        :param solution_data: optional json with an initial solution
        :return: solution, solution checks, instance checks and logs
        """
        if config.get("msg", True):
            print("Solving the model")
        validator = Draft7Validator(self.schema)
        if not validator.is_valid(config):
            error_list = [e for e in validator.iter_errors(config)]
            raise BadConfiguration(
                "The configuration does not match the schema:\n{}".format(error_list)
            )

        solver = config.get("solver")
        if solver is None:
            solver = self.get_default_solver_name()
        solver_class = self.get_solver(name=solver)
        if solver_class is None:
            raise NoSolverException("Solver {} is not available".format(solver))
        inst = self.instance.from_dict(data)
        inst_errors = inst.check_schema()
        if inst_errors:
            raise BadInstance(
                "The instance does not match the schema:\n{}".format(inst_errors)
            )
        sol = None
        if solution_data is not None:
            sol = self.solution.from_dict(solution_data)
            sol_errors = sol.check_schema()
            if sol_errors:
                raise BadSolution(
                    "The solution does not match the schema:\n{}".format(sol_errors)
                )

        instance_checks = inst.check()
        if instance_checks:
            log = dict(
                time=0,
                solver=solver,
                status="Infeasible",
                status_code=STATUS_INFEASIBLE,
                sol_code=SOLUTION_STATUS_INFEASIBLE,
            )
            return dict(), None, instance_checks, "", log

        algo = solver_class(inst, sol)
        start = timer()
        output = algo.solve(config)
        sol = None
        status_conv = {
            STATUS_OPTIMAL: "Optimal",
            STATUS_TIME_LIMIT: "Time limit",
            STATUS_INFEASIBLE: "Infeasible",
            STATUS_UNDEFINED: "Unknown",
            STATUS_NOT_SOLVED: "Not solved",
        }
        # compatibility with previous format:
        if isinstance(output, int):
            output = dict(status=output)
        status = output.get("status")
        status_sol = output.get("status_sol")
        log = dict(
            time=timer() - start,
            solver=solver,
            status=status_conv.get(status, "Unknown"),
            status_code=status,
            sol_code=SOLUTION_STATUS_INFEASIBLE,
        )
        # check if there is a solution
        # TODO: we need to extract the solution status too
        #  because there may be already an initial solution in the solver
        if status_sol is not None:
            log["sol_code"] = status_sol
        elif algo.solution is not None and len(algo.solution.data):
            log["sol_code"] = SOLUTION_STATUS_FEASIBLE

        if log["sol_code"] > 0:
            sol = algo.solution.to_dict()

        checks = algo.check_solution()

        return sol, checks, {}, "", log

    def get_solver(self, name: str = "default") -> Union[Type[ExperimentCore], None]:
        """
        :param name: name of the solver to find

        :return: the constructor for a solver matching the name
        """
        if "." in name:
            solver = name.split(".")[0]
        else:
            solver = name
        return self.solvers.get(solver)

    def get_default_solver_name(self) -> str:
        """
        :return: the name of the default solver
        """
        return self.schema["properties"]["solver"]["default"]

    def get_schemas(self) -> Dict[str, Dict]:
        """
        :return: a dictionary with the three schemas that define the solution method
        """
        return dict(
            instance=self.instance.schema,
            solution=self.solution.schema,
            config=self.schema,
        )



"""
Integration test for the Cornflow client
Base, admin and service user get tested
Integration between Airflow and cornflow through airflow client and cornflow client tested as well
"""
# Full imports
import json
import os
import pulp as pl
import time

# Partial imports
from unittest import TestCase

# Internal imports
from cornflow_client import CornFlow
from cornflow_client.constants import STATUS_OPTIMAL, STATUS_NOT_SOLVED
from cornflow_client.schema.tools import get_pulp_jsonschema
from cornflow_client.tests.const import PUBLIC_DAGS, PULP_EXAMPLE

# Constants
path_to_tests_dir = os.path.dirname(os.path.abspath(__file__))

# Helper functions
def _load_file(_file):
    with open(_file) as f:
        temp = json.load(f)
    return temp


def _get_file(relative_path):
    return os.path.join(path_to_tests_dir, relative_path)


# Testing suit classes
class TestCornflowClientUser(TestCase):
    def setUp(self):
        self.client = CornFlow(url="http://127.0.0.1:5050/")
        login_result = self.client.login("user", "UserPassword1!")
        self.assertIn("id", login_result.keys())
        self.assertIn("token", login_result.keys())
        self.user_id = login_result["id"]

    def tearDown(self):
        pass

    def test_health_endpoint(self):
        response = self.client.is_alive()
        self.assertEqual(response["cornflow_status"], "healthy")
        self.assertEqual(response["airflow_status"], "healthy")

    def test_sign_up(self):
        response = self.client.sign_up(
            "test_username", "test_username@cornflow.org", "TestPassword2!"
        )
        self.assertIn("id", response.json().keys())
        self.assertIn("token", response.json().keys())
        self.assertEqual(201, response.status_code)

    def test_create_instance(self):
        data = _load_file(PULP_EXAMPLE)
        response = self.client.create_instance(data, "test_example", "test_description")
        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "executions",
        ]
        for item in items:
            self.assertIn(item, response.keys())

        self.assertEqual("test_example", response["name"])
        self.assertEqual("solve_model_dag", response["schema"])
        self.assertEqual("test_description", response["description"])

        return response

    def test_create_case(self):
        data = _load_file(PULP_EXAMPLE)
        response = self.client.create_case(
            name="test_case",
            schema="solve_model_dag",
            data=data,
            description="test_description",
        )

        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "solution_hash",
            "path",
            "updated_at",
            "is_dir",
        ]

        for item in items:
            self.assertIn(item, response.keys())
        self.assertEqual("test_case", response["name"])
        self.assertEqual("solve_model_dag", response["schema"])
        self.assertEqual("test_description", response["description"])
        return response

    def test_create_instance_file(self):
        response = self.client.create_instance_file(
            _get_file("../data/test_mps.mps"),
            name="test_filename",
            description="filename_description",
        )

        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "executions",
        ]

        for item in items:
            self.assertIn(item, response.keys())

        self.assertEqual("test_filename", response["name"])
        self.assertEqual("solve_model_dag", response["schema"])
        self.assertEqual("filename_description", response["description"])

    def test_create_execution(self):
        instance = self.test_create_instance()
        response = self.client.create_execution(
            instance_id=instance["id"],
            config={"solver": "PULP_CBC_CMD", "timeLimit": 60},
            name="test_execution",
            description="execution_description",
            schema="solve_model_dag",
        )
        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "config",
            "instance_id",
            "state",
            "message",
        ]

        for item in items:
            self.assertIn(item, response.keys())

        self.assertEqual(instance["id"], response["instance_id"])
        self.assertEqual("test_execution", response["name"])
        self.assertEqual("execution_description", response["description"])
        self.assertEqual(
            {"solver": "PULP_CBC_CMD", "timeLimit": 60}, response["config"]
        )
        self.assertEqual(STATUS_NOT_SOLVED, response["state"])

        return response

    def test_execution_results(self):
        execution = self.test_create_execution()
        time.sleep(10)
        response = self.client.get_results(execution["id"])

        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "config",
            "instance_id",
            "state",
            "message",
        ]

        for item in items:
            self.assertIn(item, response.keys())

        self.assertEqual(execution["id"], response["id"])
        self.assertEqual(STATUS_OPTIMAL, response["state"])

    def test_execution_status(self):
        execution = self.test_create_execution()
        response = self.client.get_status(execution["id"])
        items = ["id", "state", "message", "data_hash"]
        for item in items:
            self.assertIn(item, response.keys())
        self.assertEqual(STATUS_NOT_SOLVED, response["state"])
        time.sleep(10)
        response = self.client.get_status(execution["id"])
        for item in items:
            self.assertIn(item, response.keys())
        self.assertEqual(STATUS_OPTIMAL, response["state"])

    def test_stop_execution(self):
        execution = self.test_create_execution()
        response = self.client.stop_execution(execution["id"])
        self.assertEqual(response["message"], "The execution has been stopped")

    def test_get_execution_log(self):
        execution = self.test_create_execution()
        response = self.client.get_log(execution["id"])

        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "config",
            "instance_id",
            "state",
            "message",
            "log",
        ]

        for item in items:
            self.assertIn(item, response.keys())
        self.assertEqual(execution["id"], response["id"])

    def test_get_execution_solution(self):
        execution = self.test_create_execution()
        time.sleep(10)
        response = self.client.get_solution(execution["id"])
        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "config",
            "instance_id",
            "state",
            "message",
            "data",
            "checks",
        ]

        for item in items:
            self.assertIn(item, response.keys())

        self.assertEqual(execution["id"], response["id"])
        self.assertEqual(STATUS_OPTIMAL, response["state"])

        return response

    def test_create_case_execution(self):
        execution = self.test_get_execution_solution()
        response = self.client.create_case(
            name="case_from_solution",
            schema="solve_model_dag",
            description="case_from_solution_description",
            solution=execution["data"],
        )

        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "solution_hash",
            "path",
            "updated_at",
            "is_dir",
        ]

        for item in items:
            self.assertIn(item, response.keys())

        self.assertEqual("case_from_solution", response["name"])
        self.assertEqual("solve_model_dag", response["schema"])
        self.assertEqual("case_from_solution_description", response["description"])
        return response

    def test_get_all_instances(self):
        self.test_create_instance()
        self.test_create_instance()
        instances = self.client.get_all_instances()
        self.assertGreaterEqual(len(instances), 2)

    def test_get_all_executions(self):
        self.test_stop_execution()
        self.test_stop_execution()
        executions = self.client.get_all_executions()
        self.assertGreaterEqual(len(executions), 2)

    def test_get_all_cases(self):
        self.test_create_case()
        self.test_create_case()
        cases = self.client.get_all_cases()
        self.assertGreaterEqual(len(cases), 2)

    def test_get_one_user(self):
        response = self.client.get_one_user(self.user_id)
        self.assertEqual(response.status_code, 200)

        items = ["id", "first_name", "last_name", "username", "email"]
        for item in items:
            self.assertIn(item, response.json().keys())

        self.assertEqual(self.user_id, response.json()["id"])
        self.assertEqual("user", response.json()["username"])
        self.assertEqual("user@cornflow.org", response.json()["email"])

    def test_get_one_instance(self):
        instance = self.test_create_instance()
        response = self.client.get_one_instance(instance["id"])
        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "executions",
        ]

        for item in items:
            self.assertIn(item, response.keys())
            self.assertEqual(instance[item], response[item])

    def test_get_one_case(self):
        case = self.test_create_case()
        response = self.client.get_one_case(case["id"])
        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "solution_hash",
            "path",
            "updated_at",
            "is_dir",
        ]

        for item in items:
            self.assertIn(item, response.keys())
            self.assertEqual(case[item], response[item])

    def test_get_one_case_execution(self):
        case = self.test_create_case_execution()
        response = self.client.get_one_case(case["id"])
        items = [
            "id",
            "name",
            "description",
            "created_at",
            "user_id",
            "data_hash",
            "schema",
            "solution_hash",
            "path",
            "updated_at",
            "is_dir",
        ]

        for item in items:
            self.assertIn(item, response.keys())
            self.assertEqual(case[item], response[item])

    def test_delete_one_case(self):
        case = self.test_create_case()
        response = self.client.delete_one_case(case["id"])
        self.assertEqual("The object has been deleted", response["message"])

    def test_put_one_case(self):
        case = self.test_create_case()
        response = self.client.put_one_case(case["id"], {"name": "new_case_name"})
        self.assertEqual("Updated correctly", response["message"])

    def test_patch_one_case(self):
        # TODO: Get example of data patch for the tests
        pass

    def test_delete_one_instance(self):
        instance = self.test_create_instance()
        response = self.client.delete_one_instance(instance["id"])
        self.assertEqual(200, response.status_code)
        self.assertEqual("The object has been deleted", response.json()["message"])

    def test_get_schema(self):
        response = self.client.get_schema("solve_model_dag")
        schema = {
            "config": get_pulp_jsonschema("solver_config.json"),
            "instance": get_pulp_jsonschema(),
            "solution": get_pulp_jsonschema(),
        }

        schema["config"]["properties"]["solver"]["enum"] = pl.listSolvers()
        schema["config"]["properties"]["solver"]["default"] = "PULP_CBC_CMD"

        for (key, value) in schema.items():
            self.assertDictEqual(value, response[key])

    def test_get_all_schemas(self):
        response = self.client.get_all_schemas()
        read_schemas = [dag for v in response for (_, dag) in v.items()]

        for schema in PUBLIC_DAGS:
            self.assertIn(schema, read_schemas)


class TestCornflowClientAdmin(TestCase):
    def setUp(self):
        self.client = CornFlow(url="http://127.0.0.1:5050/")
        login_result = self.client.login("admin", "Adminpassword1!")
        self.assertIn("id", login_result.keys())
        self.assertIn("token", login_result.keys())
        self.base_user_id = CornFlow(url="http://127.0.0.1:5050/").login(
            "user", "UserPassword1!"
        )["id"]

    def tearDown(self):
        pass

    def test_get_all_users(self):
        response = self.client.get_all_users()
        self.assertGreaterEqual(len(response), 3)

    def test_get_one_user(self):
        response = self.client.get_one_user(self.base_user_id)

        items = ["id", "first_name", "last_name", "username", "email"]
        for item in items:
            self.assertIn(item, response.json().keys())

        self.assertEqual(self.base_user_id, response.json()["id"])
        self.assertEqual("user", response.json()["username"])
        self.assertEqual("user@cornflow.org", response.json()["email"])


class TestCornflowClientService(TestCase):
    def setUp(self):
        self.client = CornFlow(url="http://127.0.0.1:5050/")
        login_result = self.client.login("airflow", "Airflow_test_password1")
        self.assertIn("id", login_result.keys())
        self.assertIn("token", login_result.keys())

    def tearDown(self):
        pass

    def test_get_execution_data(self):
        client = CornFlow(url="http://127.0.0.1:5050/")
        _ = client.login("user", "UserPassword1!")
        data = _load_file(PULP_EXAMPLE)
        instance = client.create_instance(data, "test_example", "test_description")
        execution = client.create_execution(
            instance_id=instance["id"],
            config={"solver": "PULP_CBC_CMD", "timeLimit": 60},
            name="test_execution",
            description="execution_description",
            schema="solve_model_dag",
        )
        response = self.client.get_data(execution["id"])
        items = ["id", "data", "config"]

        for item in items:
            self.assertIn(item, response.keys())

        self.assertEqual(instance["id"], response["id"])
        self.assertEqual(data, response["data"])
        self.assertEqual(execution["config"], response["config"])

    def test_write_execution_solution(self):
        client = CornFlow(url="http://127.0.0.1:5050/")
        _ = client.login("user", "UserPassword1!")
        data = _load_file(PULP_EXAMPLE)
        instance = client.create_instance(data, "test_example", "test_description")
        execution = client.create_execution(
            instance_id=instance["id"],
            config={"solver": "PULP_CBC_CMD", "timeLimit": 60},
            name="test_execution",
            description="execution_description",
            schema="solve_model_dag",
        )

        time.sleep(15)

        solution = client.get_solution(execution["id"])

        payload = dict(
            state=1, log_json={}, log_text="", solution_schema="solve_model_dag"
        )
        payload["data"] = solution["data"]

        response = self.client.write_solution(execution["id"], **payload)
        self.assertEqual("results successfully saved", response["message"])

    def test_get_deployed_dags(self):
        response = self.client.get_deployed_dags()

        items = ["id", "description"]
        for item in items:
            self.assertIn(item, response[0].keys())

        deployed_dags = [v["id"] for v in response]

        for dag in PUBLIC_DAGS:
            self.assertIn(dag, deployed_dags)

    def test_post_deployed_dag(self):

        response = self.client.create_deployed_dag(
            name="test_dag", description="test_dag_description"
        )

        items = ["id", "description"]

        for item in items:
            self.assertIn(item, response.keys())
        self.assertEqual("test_dag", response["id"])
        self.assertEqual("test_dag_description", response["description"])



"""

"""
# Full imports
import json
import os

# Imports from internal modules
from cornflow_client.core import InstanceSolutionCore
from cornflow_client.core.read_tools import read_excel


def get_pulp_jsonschema(filename="pulp_json_schema.json", path="data"):
    """
    returns the PuLP model schema
    """
    filename = os.path.join(os.path.dirname(__file__), "..", path, filename)
    with open(filename, "r") as f:
        content = json.load(f)
    return content


def get_empty_schema(properties=None, solvers=None):
    """
    assumes the first solver is the default
    """
    schema = get_pulp_jsonschema("empty_schema.json")
    if properties is not None:
        schema["properties"] = properties
    if solvers is not None:
        schema["properties"]["solver"] = dict(
            type="string", enum=solvers, default=solvers[0]
        )
    return schema


def schema_from_excel(path_in, param_tables=None, path_out=None):
    """
    Create a jsonschema based on an excel data file.

    :param path_in: path of the excel file
    :param param_tables: array containing the names of the parameter tables
    :param path_out: path where to save the json schema as a json file.
    :return: the jsonschema
    """
    if not param_tables:
        param_tables = []
    xl_data = read_excel(path_in, param_tables)
    data = {k: str_columns(v) if isinstance(v, list) else v for k, v in xl_data.items()}

    class InstSol(InstanceSolutionCore):
        schema = {}

    instance = InstSol(data)
    schema = instance.generate_schema()

    if path_out is not None:
        with open(path_out, "w") as f:
            json.dump(schema, f, indent=4, sort_keys=False)

    return schema


def str_key(dic):
    """
    Apply str to the keys of a dict.
    This must be a applied to a dict in order to transform it into json.

    :param dic: a dict
    :return: the dict with keys as strings.
    """
    return {str(k): v for k, v in dic.items()}


def str_columns(table):
    """
    Transform the columns of a table (the keys of a list of dict) into strings.

    :param table: a list of dict.
    :return: the modified list of dict
    """
    return [str_key(d) for d in table]



"""

"""
# Partial imports
from abc import ABC, abstractmethod
from typing import Union, Dict

# Imports from internal modules
from .instance import InstanceCore
from .solution import SolutionCore


class ExperimentCore(ABC):
    """
    The solver template.
    """

    def __init__(
        self,
        instance: InstanceCore,
        solution: Union[SolutionCore, None] = None,
    ):
        # instance is read-only
        self._instance = instance
        self.solution = solution

    @property
    def instance(self) -> InstanceCore:
        """
        :return: the instance
        """
        return self._instance

    @property
    def solution(self) -> SolutionCore:
        """
        :return: the solution
        """
        return self._solution

    @solution.setter
    def solution(self, value: SolutionCore) -> None:
        self._solution = value

    @abstractmethod
    def solve(self, options: dict) -> dict:
        """
        Mandatory method

        :param options: configuration for solving the problem
        :return: a dictionary with status codes and other information

        This method produces and stores a solution
        """
        pass

    @abstractmethod
    def get_objective(self) -> float:
        """
        Mandatory method

        :return: the value of the current solution, represented by a number
        """
        pass

    @abstractmethod
    def check_solution(self, *args, **kwargs) -> Dict[str, Dict]:
        """
        Mandatory method

        :return: a dictionary of dictionaries. Each dictionary represents one type of error. Each of the elements
          inside represents one error of that particular type.
        """
        pass



import os
from cornflow_client.constants import DATASCHEMA

path_to_tests_dir = os.path.dirname(os.path.abspath(__file__))


def _get_file(relative_path):
    return os.path.join(path_to_tests_dir, relative_path)


dict_example = dict(
    CoefficientSchema=[
        dict(name="name", type="String", required=True, allow_none=False, many=False),
        dict(name="value", type="Float", required=True, allow_none=False, many=False),
    ],
    ObjectiveSchema=[
        dict(name="name", type="String", required=False, allow_none=True, many=False),
        dict(name="coefficients", type="CoefficientSchema", many=True, required=True),
    ],
    ConstraintsSchema=[
        dict(name="name", type="String", required=True, allow_none=False, many=False),
        dict(name="sense", type="Integer", required=True, allow_none=False, many=False),
        dict(name="pi", type="Float", required=True, allow_none=True, many=False),
        dict(
            name="constant", type="Float", required=False, allow_none=True, many=False
        ),
        dict(name="coefficients", type="CoefficientSchema", many=True, required=True),
    ],
    VariablesSchema=[
        dict(name="name", type="String", required=True, allow_none=False, many=False),
        dict(
            name="lowBound", type="Float", allow_none=True, required=False, many=False
        ),
        dict(name="upBound", type="Float", allow_none=True, required=False, many=False),
        dict(name="cat", type="String", required=True, allow_none=False, many=False),
        dict(name="dj", type="Float", allow_none=True, required=False, many=False),
        dict(
            name="varValue", type="Float", allow_none=True, required=False, many=False
        ),
    ],
    ParametersSchema=[
        dict(name="name", type="String", required=True, allow_none=False, many=False),
        dict(name="sense", type="Integer", required=True, allow_none=False, many=False),
        dict(
            name="status", type="Integer", allow_none=False, many=False, required=False
        ),
        dict(
            name="sol_status",
            type="Integer",
            allow_none=False,
            many=False,
            required=False,
        ),
    ],
    Sos1Schema=[
        dict(
            name="placeholder",
            type="String",
            required=False,
            allow_none=False,
            many=False,
        ),
    ],
    Sos2Schema=[
        dict(
            name="placeholder",
            type="String",
            required=False,
            allow_none=False,
            many=False,
        ),
    ],
)

dict_example.update(
    {
        DATASCHEMA: [
            dict(name="objective", type="ObjectiveSchema", required=True, many=False),
            dict(name="parameters", type="ParametersSchema", required=True, many=False),
            dict(
                name="constraints", type="ConstraintsSchema", many=True, required=True
            ),
            dict(name="variables", type="VariablesSchema", many=True, required=True),
            dict(name="sos1", type="Sos1Schema", many=True, required=False),
            dict(name="sos2", type="Sos2Schema", many=True, required=False),
        ]
    }
)

PULP_EXAMPLE = _get_file("./data/pulp_example_data.json")

PUBLIC_DAGS = [
    "solve_model_dag",
    "graph_coloring",
    "timer",
    "bar_cutting",
    "facility_location",
    "graph_coloring",
    "hk_2020_dag",
    "knapsack",
    "roadef",
    "rostering",
    "tsp",
    "vrp",
]



"""

"""
# Partial imports

from functools import partial
from marshmallow import Schema, fields, INCLUDE

# Imports from internal modules
from cornflow_client.constants import (
    BOOLEAN_TYPE,
    STRING_TYPE,
    INTEGER_TYPE,
    FLOAT_TYPE,
    BASIC_TYPES,
)


class ParameterSchema(Schema):
    name = fields.String(required=True)
    description = fields.String(required=False)
    required = fields.Bool(default=False)
    many = fields.Bool(default=False)
    allow_none = fields.Bool(default=False)
    type = fields.String(required=True)
    valid_values = fields.List(fields.String, required=False)


def validator(valid_values, input):
    if input in valid_values:
        return True
    return False


def get_type(param_type, possible_dict=None):
    if param_type == STRING_TYPE:
        return fields.String
    if param_type == BOOLEAN_TYPE:
        return fields.Boolean
    if param_type == INTEGER_TYPE:
        return fields.Integer
    if param_type == FLOAT_TYPE:
        return fields.Float
    # we assume nested
    if possible_dict:
        return partial(fields.Nested, possible_dict[param_type])
    return None


def gen_schema(cls_name, params, possible_dict=None):
    """
    returns a marshmallow schema as if it were generated with a python object

    :param cls_name: string with the class name
    :param params: list of dictionaries
    :param possible_dict: dictionary with previously defined types
    :return:
    """
    dict_fields = {}
    # params includes at least: name, type
    for p in params:
        # p is a dictionary
        # we copy since we do not want to modify the original
        p = dict(p)
        p_type = p.pop("type")
        field_type = get_type(p_type, possible_dict=possible_dict)
        if p.get("many") and p_type in BASIC_TYPES:
            # if this is a list: we need to make an explicit list
            field_type = partial(fields.List, field_type)
        if p_type in BASIC_TYPES:
            # basic types do not support the "many" argument, apparently
            p.pop("many", None)
        valid_values = p.pop("valid_values", None)
        name = p.pop("name")
        if valid_values is not None:
            dict_fields[name] = field_type(
                validate=partial(validator, valid_values), **p
            )
        else:
            dict_fields[name] = field_type(**p)
    schema = type(cls_name, (Schema,), dict_fields)
    # HACKISH: This allows us to validate if something is missing.
    # there's probably a better way to do this:
    schema.opts.unknown = INCLUDE
    return schema


def sort_dict(dict_params):
    """
    this function returns the key, value list in an order that does not generate conflicts

    """
    backup = list(dict_params.items())
    order = dict()
    # max_count = len(order)
    count = 0
    pos = 0
    fails = 0
    check = set(BASIC_TYPES)
    while len(backup):
        key, el = backup[count]
        at_least_unseen_dependency = [ell for ell in el if ell["type"] not in check]
        if not len(at_least_unseen_dependency):
            order[key] = pos
            backup.pop(count)
            check.add(key)
            pos += 1
            fails = 0
        else:
            fails += 1
            count += 1
        if fails and fails == len(backup):
            raise ValueError(
                "sort_dict: Issue with the input dictionary."
                " There may be schemas types which are not defined"
            )
        # restart
        if count == len(backup):
            count = 0
    return order



"""

"""
# Full imports
import json
import warnings

# Partial imports
from abc import ABC, abstractmethod
from genson import SchemaBuilder
from jsonschema import Draft7Validator
from pytups import SuperDict
from typing import List

# Imports from internal modules
from .read_tools import read_excel, is_xl_type


class InstanceSolutionCore(ABC):
    """
    Common interface for the instance and solution templates
    """

    def __init__(self, data: dict):
        self.data = SuperDict.from_dict(data)

    @property
    def data(self) -> dict:
        """
        input data (not necessarily in json-format)
        """
        return self._data

    @data.setter
    def data(self, value: dict):
        self._data = value

    @classmethod
    def from_dict(cls, data: dict) -> "InstanceSolutionCore":
        """
        :param data: json-schema in a dictionary format

        :return: an object initialized from the dict json-schema
        """
        return cls(data)

    def to_dict(self) -> dict:
        """
        :return: a dictionary with the json-schema representation
        """
        return self.data

    @classmethod
    def from_json(cls, path: str) -> "InstanceSolutionCore":
        """
        :param path: path to json-schema json file

        :return: an object initialized from the json-schema formatted json file
        """
        with open(path, "r") as f:
            data_json = json.load(f)
        return cls.from_dict(data_json)

    def to_json(self, path: str) -> None:
        """
        :param path: path to json-schema json file

        writes a json file with the json-schema representation of the object
        """

        data = self.to_dict()
        with open(path, "w") as f:
            json.dump(data, f, indent=4, sort_keys=True)

    @property
    @abstractmethod
    def schema(self) -> dict:
        """
        a dictionary representation of the json-schema for the object
        """

        raise NotImplementedError()

    def check_schema(self) -> List:
        """
        checks that the json-schema export complies with the defined schema

        :return: a list of errors
        """

        validator = Draft7Validator(self.schema)
        data = self.to_dict()
        if not validator.is_valid(data):
            return [e for e in validator.iter_errors(data)]
        return []

    def generate_schema(self) -> dict:
        """
        :return: a dict json-schema based on the current data
        """
        builder = SchemaBuilder()
        builder.add_schema({"type": "object", "properties": {}})
        builder.add_object(self.to_dict())
        return builder.to_schema()

    @classmethod
    def from_excel(cls, path: str) -> "InstanceSolutionCore":
        """
        Read an entire excel file.

        :param path: path of the excel file
        :return: a dict with a list of dict (records format) for each table.
        """

        param_tables_names = cls.__get_parameter_tables_names()

        tables = read_excel(path, param_tables_names)
        return cls.from_dict(tables)

    @classmethod
    def __get_parameter_tables_names(cls) -> List:
        """
        :return: The names of the table of the schema that are parameter tables
        """
        json_schema = cls(dict()).schema
        if json_schema.get("properties", None):
            return [
                table_name
                for table_name, content in json_schema["properties"].items()
                if content["type"] == "object"
            ]
        return []

    def to_excel(self, path: str):
        """
        Write data to excel.

        :param path: path or name of the excel file
        :return: nothing
        """
        try:
            import pandas as pd
        except (ImportError, ModuleNotFoundError):
            warnings.warn("You have to install pandas to use this method")
            raise Exception("You must install pandas package to use this method")

        is_xl_type(path)

        data = self.to_dict()

        with pd.ExcelWriter(path) as writer:
            for table in data.keys():
                content = data[table]
                if isinstance(content, list):
                    pd.DataFrame.from_records(content).to_excel(
                        writer, table, index=False
                    )
                elif isinstance(content, dict):
                    pd.DataFrame.from_dict(content, orient="index").to_excel(
                        writer, table, header=False
                    )



# Import from external libraries
from datetime import datetime, timedelta


class DateExtraMethods:
    def __init__(self):
        pass

    @staticmethod
    def get_date_from_string(string: str) -> datetime:
        """
        Returns a datetime object from an hour-string in format 'YYYY-MM-DD'
        """
        return datetime.strptime(string, "%Y-%m-%d")

    @staticmethod
    def get_datetime_from_string(string: str) -> datetime:
        """
        Returns a datetime object from an hour-string in format 'YYYY-MM-DDTh:m'
        """
        return datetime.strptime(string, "%Y-%m-%dT%H:%M")

    @staticmethod
    def get_datetimesec_from_string(string: str) -> datetime:
        """
        Returns a datetime object from an hour-string in format 'YYYY-MM-DDTh:m:s'
        """
        return datetime.strptime(string, "%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def get_datetime_from_date_hour(date: str, hour: int) -> datetime:
        """
        Returns a datetime object from a date and an hour
        """
        if hour == 24:
            hour = 0
        return datetime.strptime(f"{date}T{hour}", "%Y-%m-%dT%H")

    @staticmethod
    def get_date_hour_from_string(string: str, zero_to_twenty_four=False):
        """
        Returns a tuple (date, hour) from an hour-string
        """
        date_t = DateExtraMethods.get_datetime_from_string(string)
        hour = date_t.strftime("%H")
        if hour == "00" and zero_to_twenty_four:
            hour = "24"
            date_t -= timedelta(days=1)
        date = date_t.strftime("%Y-%m-%d")
        return date, int(hour)

    @staticmethod
    def get_date_string_from_ts(ts: datetime) -> str:
        """Returns the string of a given date as 'YYYY-MM-DD'"""
        return datetime.strftime(ts, "%Y-%m-%d")

    @staticmethod
    def get_datetime_string_from_ts(ts: datetime) -> str:
        """Returns the string of a given date as 'YYYY-MM-DDTh:m'"""
        return datetime.strftime(ts, "%Y-%m-%dT%H:%M")

    @staticmethod
    def get_datetimesec_string_from_ts(ts: datetime) -> str:
        """Returns the string of a given date as 'YYYY-MM-DDTh:m:s'"""
        return datetime.strftime(ts, "%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def get_next_hour_datetime_string(string: str) -> str:
        """
        Returns the hour following the given hour, as a string
        """
        return (
            DateExtraMethods.get_datetime_from_string(string) + timedelta(hours=1)
        ).isoformat()

    @staticmethod
    def get_next_hour_datetimesec_string(string: str) -> str:
        """
        Returns the hour following the given hour, as a string
        """
        return (
            DateExtraMethods.get_datetimesec_from_string(string) + timedelta(hours=1)
        ).isoformat()

    @staticmethod
    def get_next_hour(ts: datetime) -> datetime:
        """
        Returns the hour following the given hour
        """
        return ts + timedelta(hours=1)

    @staticmethod
    def get_previous_hour_datetime_string(string: str) -> str:
        """
        Returns the hour preceding the given hour, as a string
        """
        return (
            DateExtraMethods.get_datetime_from_string(string) - timedelta(hours=1)
        ).isoformat()

    @staticmethod
    def get_previous_hour_datetimesec_string(string: str) -> str:
        """
        Returns the hour preceding the given hour, as a string
        """
        return (
            DateExtraMethods.get_datetimesec_from_string(string) - timedelta(hours=1)
        ).isoformat()

    @staticmethod
    def get_previous_hour(ts: datetime) -> datetime:
        """
        Returns the hour preceding the given hour
        """
        return ts - timedelta(hours=1)

    @staticmethod
    def get_date_string_from_ts_string(ts: str) -> str:
        """Returns the date in format 'YYYY-MM-DD' from a datetime string"""
        return ts[0:10]

    @staticmethod
    def get_hour_from_ts(ts: datetime) -> float:
        """Returns the hours (in number) of the given time slot"""
        return float(ts.hour + ts.minute / 60)

    @staticmethod
    def add_time_to_ts(ts: datetime, weeks=0, days=0, minutes=0, seconds=0) -> datetime:
        """Adds time to a datetime"""
        return ts + timedelta(days=7 * weeks + days, minutes=minutes, seconds=seconds)

    @staticmethod
    def add_time_to_date_string(
        string: str, weeks=0, days=0, minutes=0, seconds=0
    ) -> str:
        """Adds time to a datetime"""
        return DateExtraMethods.get_date_string_from_ts(
            DateExtraMethods.get_date_from_string(string)
            + timedelta(days=7 * weeks + days, minutes=minutes, seconds=seconds)
        )

    @staticmethod
    def add_time_to_datetime_string(
            string: str, weeks=0, days=0, minutes=0, seconds=0
    ) -> str:
        """Adds time to a datetime"""
        return DateExtraMethods.get_datetime_string_from_ts(
            DateExtraMethods.get_datetime_from_string(string)
            + timedelta(days=7 * weeks + days, minutes=minutes, seconds=seconds)
        )

    @staticmethod
    def add_time_to_datetimesec_string(
        string: str, weeks=0, days=0, hours=0, minutes=0, seconds=0
    ) -> str:
        """Adds time to a datetime"""
        return DateExtraMethods.get_datetimesec_string_from_ts(
            DateExtraMethods.get_datetimesec_from_string(string)
            + timedelta(days=7 * weeks + days, hours=hours, minutes=minutes, seconds=seconds)
        )

    @staticmethod
    def get_week_from_ts(ts: datetime) -> int:
        """Returns the integer value of the week for the given time slot"""
        return ts.isocalendar()[1]

    @staticmethod
    def get_week_from_date_string(string: str) -> int:
        """Returns the integer value of the week for the given string"""
        return DateExtraMethods.get_week_from_ts(
            DateExtraMethods.get_date_from_string(string)
        )

    @staticmethod
    def get_week_from_datetime_string(string: str) -> int:
        """Returns the integer value of the week for the given string"""
        return DateExtraMethods.get_week_from_ts(
            DateExtraMethods.get_datetime_from_string(string)
        )

    @staticmethod
    def get_week_from_datetimesec_string(string: str) -> int:
        """Returns the integer value of the week for the given string"""
        return DateExtraMethods.get_week_from_ts(
            DateExtraMethods.get_datetimesec_from_string(string)
        )

    @staticmethod
    def get_weekday_from_ts(ts: datetime) -> int:
        """ Returns the number of the weekday from a ts """
        return ts.isocalendar()[2]

    @staticmethod
    def get_weekday_from_date_string(string: str) -> int:
        """ Returns the number of the weekday from a date string in format 'YYYY-MM-DD' """
        return DateExtraMethods.get_date_from_string(string).isocalendar()[2]

    @staticmethod
    def get_weekday_from_datetime_string(string: str) -> int:
        """ Returns the number of the weekday from a date string in format 'YYYY-MM-DDTh:m' """
        return DateExtraMethods.get_datetime_from_string(string).isocalendar()[2]

    @staticmethod
    def get_weekday_from_datetimesec_string(string: str) -> int:
        """ Returns the number of the weekday from a date string in format 'YYYY-MM-DDT:h:m:s' """
        return DateExtraMethods.get_datetimesec_from_string(string).isocalendar()[2]

    @staticmethod
    def get_hour_from_datetime_string(string: str) -> float:
        """Returns the integer value of the hour (in number) from ts string in format 'YYYY-MM-DDTh:m'"""
        return DateExtraMethods.get_hour_from_ts(
            DateExtraMethods.get_datetime_from_string(string)
        )

    @staticmethod
    def get_hour_from_datetimesec_string(string: str) -> float:
        """Returns the integer value of the hour (in number) from ts string in format 'YYYY-MM-DDTh:m:s'"""
        return DateExtraMethods.get_hour_from_ts(
            DateExtraMethods.get_datetimesec_from_string(string)
        )


class ExtraMethods(DateExtraMethods):
    def __init__(self):
        DateExtraMethods.__init__(self)



"""

"""
# Full imports
import json
import os

# Partial imports
from datetime import datetime, timedelta
from urllib.parse import urlparse, urljoin


# Imports from modules
from cornflow_client import CornFlow, CornFlowApiError


# TODO: convert everything to an object that encapsulates everything
#  to make it clear and avoid all the arguments.
# class DagApp(object):
#
#     def __init__(self, dag_name, secrets):
#         self.name = dag_name
#         self.secrets = secrets
#
#     def solve(self, data, config):
#         raise NotImplemented()
#
#     def test_cases(self):
#         raise NotImplemented()
#
#     pass


default_args = {
    "owner": "baobab",
    "depends_on_past": False,
    "start_date": datetime(2020, 2, 1),
    "email": [""],
    "email_on_failure": False,
    "email_on_retry": False,
    "retry_delay": timedelta(minutes=1),
    "schedule_interval": None,
}


def get_schemas_from_file(_dir, dag_name):
    with open(os.path.join(_dir, dag_name + "_input.json"), "r") as f:
        instance = json.load(f)
    with open(os.path.join(_dir, dag_name + "_output.json"), "r") as f:
        solution = json.load(f)
    return instance, solution


def get_requirements(path):
    """
    Read requirements.txt from a project and return a list of packages.

    :param path: The path of the project
    :return: A list of required packages
    """
    req_path = path + "/requirements.txt"

    try:
        with open(req_path, "r") as file:
            req_list = file.read().splitlines()
    except:
        print("no requirement file in this path")
        return []

    return req_list


def connect_to_cornflow(secrets):
    """
    Create a connection to cornflow and log in with airflow admin user.

    :return: A logged and connected Cornflow class instance
    """
    # This secret comes from airflow configuration
    print("Getting connection information from ENV VAR=CF_URI")
    uri = secrets.get_conn_uri("CF_URI")
    conn = urlparse(uri)
    scheme = conn.scheme
    if scheme == "cornflow":
        scheme = "http"
    url = "{}://{}".format(scheme, conn.hostname)
    if conn.port:
        url += ":{uri.port}".format(uri=conn)
    if conn.path:
        url = urljoin(url, conn.path)
    airflow_user = CornFlow(url=url)
    airflow_user.login(username=conn.username, pwd=conn.password)
    return airflow_user


def try_to_save_error(client, exec_id, state=-1):
    """
    Attempt at saving that the execution failed
    """
    try:
        client.put_api_for_id("dag/", id=exec_id, payload=dict(state=state))
    except Exception as e:
        print("An exception trying to register the failed status: {}".format(e))


def try_to_write_solution(client, exec_id, payload):
    """
    Tries to write the payload into cornflow
    If it fails tries to write again that it failed.
    If it fails at least once: it raises an exception
    """
    execution_payload = dict(**payload)
    execution_payload.pop("inst_id")
    execution_payload.pop("inst_checks")
    try:
        client.write_solution(execution_id=exec_id, **execution_payload)
    except CornFlowApiError:
        try_to_save_error(client, exec_id, -6)
        # attempt to update the execution with a failed status.
        raise AirflowDagException("The writing of the solution failed")

    if payload["inst_checks"]:
        checks_payload = dict()
        checks_payload["checks"] = payload["inst_checks"]
        try:
            client.write_instance_checks(
                instance_id=payload["inst_id"], **checks_payload
            )
        except CornFlowApiError:
            try_to_save_error(client, exec_id, -6)
            raise AirflowDagException("The writing of the instance checks failed")


def get_schema(dag_name):
    _file = os.path.join(os.path.dirname(__file__), "{}_output.json".format(dag_name))
    with open(_file, "r") as f:
        schema = json.load(f)
    return schema


def cf_solve_app(app, secrets, **kwargs):
    return cf_solve(app.solve, app.name, secrets, **kwargs)


def cf_solve(fun, dag_name, secrets, **kwargs):
    """
    Connect to cornflow, ask for data, solve the problem and write the solution in cornflow

    :param fun: The function to use to solve the problem.
    :param dag_name: the name of the dag, to later search the output schema
    :param kwargs: other kwargs passed to the dag task.
    :return:
    """
    client = connect_to_cornflow(secrets)
    exec_id = kwargs["dag_run"].conf["exec_id"]
    execution_data = client.get_data(exec_id)
    data = execution_data["data"]
    config = execution_data["config"]
    inst_id = execution_data["id"]
    try:
        solution, sol_checks, inst_checks, log, log_json = fun(data, config)
    except NoSolverException as e:
        if config.get("msg", True):
            print("No solver found !")
        try_to_save_error(client, exec_id, -1)
        raise AirflowDagException(e)
    except Exception as e:
        if config.get("msg", True):
            print("Some unknown error happened")
        try_to_save_error(client, exec_id, -1)
        raise AirflowDagException("There was an error during the solving")

    payload = dict(
        state=1,
        log_json=log_json,
        log_text=log,
        solution_schema=dag_name,
        inst_checks=inst_checks,
        inst_id=inst_id,
    )

    if not solution:
        # No solution found: we just send everything to cornflow.
        if config.get("msg", True):
            print("No solution found: we save what we have.")
        try_to_write_solution(client, exec_id, payload)
        return "Solution was not saved"
    # There is a solution:
    # we first need to validate the schema.
    # If it's not: we change the status to Invalid
    # and take out the server validation of the schema\
    payload["data"] = solution

    if config.get("msg", True):
        print("A solution was found: we will first validate it")

    if sol_checks is not None:
        payload["checks"] = sol_checks

    try_to_write_solution(client, exec_id, payload)

    # The validation went correctly: can save the solution without problem
    return "Solution saved"


class NoSolverException(Exception):
    pass


class AirflowDagException(Exception):
    pass



import unittest
from cornflow_client.airflow import dag_utilities as du
from unittest.mock import Mock, patch


class DagUtilities(unittest.TestCase):
    @patch("cornflow_client.airflow.dag_utilities.CornFlow")
    def test_env_connection_vars(self, CornFlow):
        secrets = Mock()
        conn_uris = [
            (
                "cornflow://some_test_user:very_classified_password@devsm.cornflow.baobabsoluciones.app",
                ("some_test_user", "very_classified_password"),
                "http://devsm.cornflow.baobabsoluciones.app",
            ),
            (
                "https://some_test_user:very_classified_password@devsm.cornflow.baobabsoluciones.app",
                ("some_test_user", "very_classified_password"),
                "https://devsm.cornflow.baobabsoluciones.app",
            ),
            (
                "https://some_test_user:very_classified_password@devsm.cornflow.baobabsoluciones.app/some_dir",
                ("some_test_user", "very_classified_password"),
                "https://devsm.cornflow.baobabsoluciones.app/some_dir",
            ),
            (
                "http://airflow:airflow_test_password@localhost:5000",
                ("airflow", "airflow_test_password"),
                "http://localhost:5000",
            ),
        ]
        client_instance = CornFlow.return_value
        client_instance.login.return_value = ""
        for (conn_str, user_info, url) in conn_uris:
            secrets.get_conn_uri.return_value = conn_str
            du.connect_to_cornflow(secrets)
            client_instance.login.assert_called_with(
                username=user_info[0], pwd=user_info[1]
            )
            CornFlow.assert_called_with(url=url)



"""

"""
import json
import os
import pickle
from unittest import TestCase

from cornflow_client import InstanceCore

from cornflow_client.schema.tools import get_empty_schema


# Constants
path_to_tests_dir = os.path.dirname(os.path.abspath(__file__))

# Helper functions
def _load_file(_file):
    with open(_file) as f:
        temp = json.load(f)
    return temp


def _get_file(relative_path):
    return os.path.join(path_to_tests_dir, relative_path)


class TestSimpleApplicationDag(TestCase):
    def setUp(self):
        self.class_to_use = SimpleInstance
        self.instance_filename = _get_file("../data/gc_input.json")
        self.instance_data = _load_file(self.instance_filename)
        self.schema = get_empty_schema()

    def tearDown(self):
        pass

    def test_instance_from_dict(self):
        instance = self.class_to_use.from_dict(self.instance_data)
        self.assertEqual(self.instance_data, instance.data)
        return instance

    def test_instance_to_dict(self):
        instance = self.test_instance_from_dict()
        self.assertEqual(self.instance_data, instance.to_dict())

    def test_instance_from_json(self):
        instance = self.class_to_use.from_json(self.instance_filename)
        self.assertEqual(self.instance_data, instance.data)
        return instance

    def test_instance_to_json(self):
        instance = self.test_instance_from_json()
        instance.to_json(_get_file("../data/gc_input_export.json"))
        read_data = _load_file(_get_file("../data/gc_input.json"))
        self.assertEqual(self.instance_data, read_data)
        os.remove(_get_file("../data/gc_input_export.json"))

    def test_instance_schema(self):
        instance = self.test_instance_from_dict()
        self.assertEqual(self.schema, instance.schema)

    def test_schema_validation(self):
        instance = self.test_instance_from_dict()
        self.assertEqual([], instance.check_schema())

    def test_schema_generation(self):
        instance = self.test_instance_from_dict()
        schema = _load_file(_get_file("../data/graph_coloring_input.json"))
        schema["properties"]["pairs"].pop("description")
        self.assertEqual(schema, instance.generate_schema())

    def test_to_excel(self):
        instance = self.test_instance_from_dict()
        instance.to_excel(_get_file("../data/gc_input_export.xlsx"))
        instance_2 = self.class_to_use.from_excel(
            _get_file("../data/gc_input_export.xlsx")
        )
        self.assertEqual(instance.data, instance_2.data)
        os.remove(_get_file("../data/gc_input_export.xlsx"))


class TestCustomInstanceDag(TestSimpleApplicationDag):
    def setUp(self):
        super().setUp()
        self.class_to_use = Instance
        self.schema = _load_file(_get_file("../data/graph_coloring_input.json"))

    def test_instance_from_dict(self):
        """Here as to be not equal as the data has been processed on the from_dict method"""
        instance = self.class_to_use.from_dict(self.instance_data)
        self.assertNotEqual(self.instance_data, instance.data)
        return instance

    def test_instance_from_json(self):
        """Here as to be not equal as the data has been processed on the from_dict method"""
        instance = self.class_to_use.from_json(self.instance_filename)
        self.assertNotEqual(self.instance_data, instance.data)
        return instance


class SimpleInstance(InstanceCore):
    schema = get_empty_schema()


class Instance(InstanceCore):
    schema = _load_file(_get_file("../data/graph_coloring_input.json"))

    @classmethod
    def from_dict(cls, data):
        tables = ["pairs"]

        data_p = {el: {(v["n1"], v["n2"]): v for v in data[el]} for el in tables}

        return cls(data_p)

    def to_dict(self):
        tables = ["pairs"]

        data_p = {el: self.data[el].values_l() for el in tables}
        return pickle.loads(pickle.dumps(data_p, -1))



"""

"""
# Full imports
import math
import warnings

# Partial imports
from pytups import TupList, SuperDict
import numbers


def read_excel(path: str, param_tables_names: list = None) -> dict:
    """
    Read an entire excel file.

    :param path: path of the excel file
    :param param_tables_names: names of the parameter tables
    :return: a dict with a list of dict (records format) for each table.
    """
    is_xl_type(path)

    try:
        import openpyxl
    except (ModuleNotFoundError, ImportError) as e:
        warnings.warn("You must install openpyxl package to use this method")
        raise Exception("You must install openpyxl package to use this method")

    try:
        import pandas as pd
    except (ModuleNotFoundError, ImportError):
        warnings.warn("You must install pandas package to use this method")
        raise Exception("You must install pandas package to use this method")

    data = pd.read_excel(path, sheet_name=None)

    data_tables = {
        name: TupList(content.to_dict(orient="records")).vapply(
            lambda v: SuperDict(v).vapply(lambda vv: format_value(vv))
        )
        for name, content in data.items()
        if name not in param_tables_names
    }

    parameters_tables = {
        t: SuperDict(read_param_table(path, t)).vapply(lambda v: format_value(v))
        for t in param_tables_names
    }

    return {**data_tables, **parameters_tables}


def read_param_table(path: str, table: str) -> dict:
    """
    Read a list of parameters and their values from excel as a dict.

    :param path: the excel file path
    :param table: the table name
    :return: a dict {param1: val1}
    """

    content = read_excel_table(path, table, header=None)
    return {d[0]: d[1] for d in content}


def read_excel_table(path: str, table: str, **kwargs):
    """
    Read a table from excel

    :param path: the excel file path
    :param table: the table name
    :return: a list of dict
    """
    try:
        import pandas as pd
    except (ModuleNotFoundError, ImportError):
        warnings.warn("You must install pandas package to use this method")
        raise Exception("You must install pandas package to use this method")

    data = pd.read_excel(path, sheet_name=table, **kwargs)

    return data.to_dict(orient="records")


def format_value(value):
    """"""
    if isinstance(value, str):
        if value in ["TRUE", "True"]:
            return True
        if value in ["FALSE", "False"]:
            return False
        return value
    if isinstance(value, bool):
        return value

    if not isinstance(value, numbers.Number):
        return str(value)
    if math.isnan(value):
        return None
    return value


def is_xl_type(path: str):
    """
    Check if a path is an excel file. Raises an error if the file is not an excel file

    :param path: path of the file
    """
    if not any(ext in path for ext in [".xlsx", ".xls", ".xlsm"]):
        raise ConnectionError(
            "Excel_file argument should be a string with an excel extension"
        )



"""

"""
# Full imports
import re

# Imports from internal modules
from .dict_functions import gen_schema, ParameterSchema, sort_dict
from cornflow_client.constants import JSON_TYPES, DATASCHEMA


class DictSchema:
    """
    A json-schema to dict-schema parser
    """

    def __init__(self, jsonschema):
        """
        Class to manage internal dictionary schema

        :param jsonschema: a json schema
        """
        self.types = JSON_TYPES
        schema_dict = self.get_empty_schema()

        if "definitions" in jsonschema:
            for item in jsonschema["definitions"].items():
                self._get_element_dict(schema_dict=schema_dict, item=item)

        if "properties" in jsonschema:
            for item in jsonschema["properties"].items():
                self._get_element_dict(schema_dict=schema_dict, item=item)
                self._create_data_schema(
                    schema_dict=schema_dict,
                    item=item,
                    required_list=jsonschema.get("required"),
                )
        self.schema = schema_dict

    def get_schema(self):
        return self.schema

    @staticmethod
    def get_empty_schema():
        """
        Create un empty schema dict
        """
        return {DATASCHEMA: []}

    def _create_data_schema(self, schema_dict, item, required_list=None):
        """
        Add a schema to schema_dict[DATASCHEMA]

        :param item: (key, value) of a dict. The key contains the name of the schema
            and the value contains its content.

        return the schema dict.
        """
        name, content = item
        if required_list is None:
            required_list = []

        schema = dict(
            name=name,
            type=self._get_type_or_new_schema(item),
            many=("type" in content and content["type"] == "array"),
            required=name in required_list,
        )
        schema_dict[DATASCHEMA].append(schema)

        return schema

    def _get_element_dict(self, schema_dict, item, required_list=None):
        """
        Parse an item (key, value) from the jsonschema and return the corresponding dict.

        :param item: An item from the jsonschema (key, value)
        :param required_list: A list of names corresponding to the required fields in the parent object

        :return A dict element for a schema_dict.
        """
        if required_list is None:
            required_list = []

        name, content = item
        if "type" not in content:
            if "$ref" in content:
                return {
                    "name": name,
                    "type": self._get_ref(item),
                    "many": False,
                    "required": (name in required_list),
                }
            else:
                print("\nType missing for item: {}".format(name))
                raise TypeError("Type missing")

        if content["type"] == "object":
            return {
                "name": name,
                "type": self._get_object_schema(schema_dict=schema_dict, item=item),
                "many": False,
                "required": (name in required_list),
            }
        elif content["type"] == "array":
            return {
                "name": name,
                "type": self._get_array_schema(schema_dict=schema_dict, item=item),
                "many": True,
                "required": (name in required_list),
            }
        else:
            return self._get_field_dict(item, required_list)

    def _get_object_schema(self, schema_dict, item):
        """
        Transform an object item from the jsonschema in a dict for the schema_dict and update self.schema_dict.
        In jsonschema objects are similar to python dict.
        The object in jsonschema is in the following format:
        "object_name": {"type":"object", "properties":{"field1": {...}, "filed2": {...}}, "required": ["field1]}

        The schema_dict object use the format:
        {"schema_name": [{"name":"field1", "type": "field1_type", "many": False, "required":(True or False)}, ...]

        :param item: The jsonschema item (key, value)
        The format of the item is: ("object_name", {"type":"object", "properties":{"a": {...}, "b": {...}})

        :return: The schema name
        """
        name, content = item
        schema_name = self._get_new_schema_name(schema_dict=schema_dict, name=name)
        ell = {
            schema_name: [
                self._get_element_dict(
                    schema_dict=schema_dict,
                    item=i,
                    required_list=self._get_required(content),
                )
                for i in content["properties"].items()
            ]
        }
        schema_dict.update(ell)
        return schema_name

    def _get_array_schema(self, schema_dict, item):
        """
        Transform a array item from the jsonschema in a dict for the schema_dict and update self.schema_dict.
        In jsonschema arrays are similar to python lists.
        The object in jsonschema is in the following format:
        "object_name": {"type":"array", "items":{format_of_items}}

        The schema_dict object use the format:
        {"schema_name": [{"name":"field1", "type": "field1_type", "many": False, "required":(True or False)

        :param item: The jsonschema item (key, value)
        The format of the item is: ("object_name", {"type":"object", "properties":{"a": {...}, "b": {...}})

        :return: The schema name
        """
        name, content = item
        content = content["items"]
        schema_name = self._get_new_schema_name(schema_dict=schema_dict, name=name)
        if "type" in content and content["type"] == "object":
            schema_dict.update(
                {
                    schema_name: [
                        self._get_element_dict(
                            schema_dict=schema_dict,
                            item=i,
                            required_list=self._get_required(content),
                        )
                        for i in content["properties"].items()
                    ]
                }
            )
        elif "$ref" in content:
            schema_name = self._get_ref((None, content))
        elif "type" in content and content["type"] != "array":
            return self._get_type(content["type"])
        else:
            schema_dict.update(
                {
                    schema_name: [
                        self._get_element_dict(
                            schema_dict=schema_dict,
                            item=i,
                            required_list=self._get_required(content),
                        )
                        for i in content.items()
                    ]
                }
            )
        return schema_name

    def _get_field_dict(self, item, required_list=None):
        """
        Transform a "normal" item from the jsonschema in a dict for the schema_dict and return it.
        This is used for items that will directly translate into fields.

        :param item: The jsonschema item in format (key, value)
        :param required_list: a list of the fields required in the parent object.

        :return: the schema_dict for this item
        """

        d = dict(
            name=item[0],
            type=self._get_type(item[1]["type"]),
            required=(item[0] in required_list),
            allow_none=("null" in item[1]["type"]),
            many=False,
        )
        return d

    def _get_ref(self, item):
        """
        Get the name of the schema for a jsonschema reference.
        jsonschema definitions are parsed first and corresponding schema are created so a schema should exist
         corresponding to the reference.

        :param item: The jsonschema item in format (key, value)
        The value should be in the following format: {"$ref": "#/definitions/object_name"}

        :return The schema name (_get_schema_name(object_name))
        """
        content = item[1]
        ref = re.search("definitions/(.+)", content["$ref"]).group(1)
        return self._get_schema_name(ref)

    def _get_type_or_new_schema(self, item):
        """
        returns a new schema or a type depending on the json_type
        """
        name, content = item
        if "type" not in content or content["type"] == "object":
            return self._get_schema_name(name)
        elif content["type"] == "array":
            return self._get_type_or_new_schema((name, content["items"]))
        else:
            return self._get_type(content["type"])

    def _get_type(self, json_type):
        """
        Translate the type between jsonschema and schema_dict.

        :param json_type: the type in jsonschema

        :return: the type in schema_dict.
        """
        if type(json_type) is list:
            not_null_type = [i for i in json_type if i != "null"]
            if len(not_null_type) > 1:
                raise Warning("Warning: more than one type given")
            return self.types[not_null_type[0]]
        else:
            return self.types[json_type]

    @staticmethod
    def _get_schema_name(name, n=0):
        """
        Transform an element name into a schema name in order to create a schema corresponding to an object or array.
        The schema name use the following format:
        [name][n]Schema (for example if name is "values" and n is 3: Values3Schema)

        :param name: The name of the object or array.
        :param n: if n is different from 0, it is added to the schema name.

        :return: the corresponding schema name.
        """
        if n == 0:
            return name.capitalize() + "Schema"
        else:
            return name.capitalize() + str(n) + "Schema"

    def _get_new_schema_name(self, schema_dict, name, n=0):
        try_name = self._get_schema_name(name, n)

        if try_name in schema_dict:
            return self._get_new_schema_name(
                schema_dict=schema_dict, name=name, n=n + 1
            )
        else:
            return try_name

    @staticmethod
    def _get_required(content):
        """
        Get the list of required name of it exist.

        :content: the dict which should have a "required" key.value

        :return: The required list or empty list.

        """
        return content.get("required", [])

    def to_marshmallow(self):
        dict_params = self.schema
        result_dict = {}
        ordered = sort_dict(dict_params)
        tuplist = sorted(dict_params.items(), key=lambda v: ordered[v[0]])
        for key, params in tuplist:
            schema = ParameterSchema()
            # this line validates the list of parameters:
            params1 = schema.load(params, many=True)
            result_dict[key] = gen_schema(key, params1, result_dict)

        return result_dict[DATASCHEMA]



"""
Constants values used in schemas functions.
"""
import pulp as pl
from ortools.sat.python import cp_model


STRING_TYPE = "String"
BOOLEAN_TYPE = "Boolean"
INTEGER_TYPE = "Integer"
FLOAT_TYPE = "Float"
BASIC_TYPES = [STRING_TYPE, BOOLEAN_TYPE, INTEGER_TYPE, FLOAT_TYPE]

JSON_TYPES = {
    "string": STRING_TYPE,
    "number": FLOAT_TYPE,
    "integer": INTEGER_TYPE,
    "null": None,
    "boolean": BOOLEAN_TYPE,
}

DATASCHEMA = "DataSchema"

INSTANCE_SCHEMA = "instance"
SOLUTION_SCHEMA = "solution"
CONFIG_SCHEMA = "config"

# why it stopped
STATUS_NOT_SOLVED = 0
STATUS_OPTIMAL = 1
STATUS_INFEASIBLE = -1
STATUS_UNBOUNDED = -2
STATUS_UNDEFINED = -3
STATUS_FEASIBLE = 2
STATUS_MEMORY_LIMIT = 3
STATUS_NODE_LIMIT = 4
STATUS_TIME_LIMIT = -4
STATUS_LICENSING_PROBLEM = -5

# is there a solution?
SOLUTION_STATUS_INFEASIBLE = 0
SOLUTION_STATUS_FEASIBLE = 2

PYOMO_STOP_MAPPING = {
    "unbounded": STATUS_UNBOUNDED,
    "infeasible": STATUS_INFEASIBLE,
    "invalidProblem": STATUS_NOT_SOLVED,
    "solverFailure": STATUS_NOT_SOLVED,
    "internalSolverError": STATUS_NOT_SOLVED,
    "error": STATUS_NOT_SOLVED,
    "userInterrupt": STATUS_NOT_SOLVED,
    "resourceInterrupt": STATUS_NOT_SOLVED,
    "licensingProblem": STATUS_LICENSING_PROBLEM,
    "maxTimeLimit": STATUS_TIME_LIMIT,
    "maxIterations": STATUS_TIME_LIMIT,
    "maxEvaluations": STATUS_TIME_LIMIT,
    "globallyOptimal": STATUS_OPTIMAL,
    "locallyOptimal": STATUS_OPTIMAL,
    "optimal": STATUS_OPTIMAL,
    "minFunctionValue": STATUS_UNDEFINED,
    "minStepLength": STATUS_UNDEFINED,
    "other": STATUS_UNDEFINED,
}

PYOMO_STATUS_MAPPING = {
    "ok": SOLUTION_STATUS_FEASIBLE,
    "warning": SOLUTION_STATUS_FEASIBLE,
    "error": SOLUTION_STATUS_INFEASIBLE,
    "aborted": SOLUTION_STATUS_INFEASIBLE,
    "unknown": SOLUTION_STATUS_INFEASIBLE,
}

ORTOOLS_STATUS_MAPPING = {
    cp_model.OPTIMAL: STATUS_OPTIMAL,
    cp_model.FEASIBLE: STATUS_FEASIBLE,
    cp_model.INFEASIBLE: STATUS_INFEASIBLE,
    cp_model.UNKNOWN: STATUS_UNDEFINED,
    cp_model.MODEL_INVALID: STATUS_UNDEFINED,
}

PULP_STATUS_MAPPING = {
    pl.LpStatusInfeasible: STATUS_INFEASIBLE,
    pl.LpStatusNotSolved: STATUS_NOT_SOLVED,
    pl.LpStatusOptimal: STATUS_OPTIMAL,
    pl.LpStatusUnbounded: STATUS_UNBOUNDED,
    pl.LpStatusUndefined: STATUS_UNDEFINED,
}


class InvalidUsage(Exception):
    status_code = 400
    error = "Unknown error"

    def __init__(self, error=None, status_code=None, payload=None):
        Exception.__init__(self)
        if error is not None:
            self.error = error
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv["error"] = self.error
        return rv


class AirflowError(InvalidUsage):
    status_code = 400

    def __init__(self, error, status_code=None, payload=None):
        self.error = error
        self.payload = payload
        if status_code is not None:
            self.status_code = status_code


class NoSolverException(Exception):
    pass


class BadConfiguration(Exception):
    pass


class BadInstance(Exception):
    pass


class BadSolution(Exception):
    pass



"""
Class to help create and manage data schema and to validate json files.
"""
# Full imports
import os

# Partial imports
from copy import deepcopy
from genson import SchemaBuilder
from jsonschema import Draft7Validator


# Imports form internal modules
from .dictSchema import DictSchema
from cornflow_client.core.tools import load_json, save_json


class SchemaManager:
    """
    A schema manager between json-schema, dict-schema and marshmallow
    """

    def __init__(self, schema, validator=Draft7Validator):
        """
        Class to help create and manage data schema.
        Once a schema is loaded, allow the validation of data.

        :param schema: a json schema
        """
        self.validator = validator
        self.jsonschema = schema

    @classmethod
    def from_filepath(cls, path):
        """
        Load a json schema from a json file.

        :param path the file path

        return The SchemaManager instance
        """

        schema = cls.load_json(path)
        return cls(schema)

    def get_jsonschema(self):
        """
        Return a copy of the stored jsonschema.
        """
        return deepcopy(self.jsonschema)

    def get_validation_errors(self, data):
        """
        Validate json data according to the loaded jsonschema and return a list of errors.
        Return an empty list if data is valid.

        :param dict data: data to validate.

        :return: A list of validation errors.

        For more details about the error format, see:
        https://python-jsonschema.readthedocs.io/en/latest/errors/#jsonschema.exceptions.ValidationError
        """
        v = self.validator(self.get_jsonschema())

        if not v.is_valid(data):
            error_list = [e for e in v.iter_errors(data)]
            return error_list
        return []

    def validate_data(self, data, print_errors=False):
        """
        Validate json data according to the loaded jsonschema.

        :param dict data: the data to validate.
        :param bool print_errors: If true, will print the errors.

        :return: True if data format is valid, else False.
        """
        errors_list = self.get_validation_errors(data)

        if print_errors:
            for e in errors_list:
                print(e)

        return len(errors_list) == 0

    def validate_schema(self, print_errors=False):
        """
        Validate the loaded jsonschema
        :param bool print_errors: If true, will print the errors

        :return: True if jsonschema is valid, else False
        """
        path_schema_validator = os.path.join(
            os.path.dirname(__file__), "../data/schema_validator.json"
        )
        validation_schema = load_json(path_schema_validator)
        v = self.validator(validation_schema)

        if v.is_valid(self.get_jsonschema()):
            return True

        error_list = [e for e in v.iter_errors(self.get_jsonschema())]
        if print_errors:
            for e in error_list:
                print(e)
        return False

    def get_file_errors(self, path):
        """
        Get json file errors according to the loaded jsonschema.

        :param path the file path

        :return: A list of validation errors.
          For more details about the error format, see:
          https://python-jsonschema.readthedocs.io/en/latest/errors/#jsonschema.exceptions.ValidationError
        """
        data = self.load_json(path)
        return self.get_validation_errors(data)

    def validate_file(self, path, print_errors=False):
        """
        Validate a json file according to the loaded jsonschema.

        :param path the file path
        :param print_errors: If true, will print the errors.

        :return: True if the data is valid and False if it is not.
        """
        data = self.load_json(path)
        return self.validate_data(data, print_errors=print_errors)

    def to_dict_schema(self):
        """
        Transform a jsonschema into a dictionary format

        :return: The schema dictionary
        """

        return self.to_schema_dict_obj().get_schema()

    def to_schema_dict_obj(self):
        """
        Returns an DictSchema object equivalent of the jsonschema

        """
        return DictSchema(self.get_jsonschema())

    @property
    def schema_dict(self):
        return self.to_dict_schema()

    def to_marshmallow(self):
        """
        Create marshmallow schemas

        :return: a dict containing the flask marshmallow schemas
        :rtype: Schema()
        """
        return self.to_schema_dict_obj().to_marshmallow()

    def export_schema_dict(self, path):
        """
        Print the schema_dict in a json file.

        :param path: the path where to save the dict.format

        :return: nothing
        """
        self.save_json(self.to_dict_schema(), path)

    def draft_schema_from(self, path, save_path=None):
        """
        Create a draft jsonschema from a json file of data.

        :param path: path to the json file.
        :param save_path: path where to save the generated schema.

        :return: the generated schema.
        """
        file = self.load_json(path)

        builder = SchemaBuilder()
        builder.add_schema({"type": "object", "properties": {}})
        builder.add_object(file)

        draft_schema = builder.to_json()
        if save_path is not None:
            with open(save_path, "w") as outfile:
                outfile.write(draft_schema)
        return draft_schema

    def to_template(self):
        """

        This function assumes certain structure for the jsonschema.
        For now, three types of tables exist: array of objects, arrays and objects.
        {
        table1: [{col1: a, col2: b}, {col1: aa, col2: bb}, ...],
        table2: [1, 2, 3, ],
        table3: {config1: a, config2: b},
        }

        """
        master_table_name = "_README"
        type_table_name = "_TYPES"
        tables = {master_table_name: [], type_table_name: []}
        # we update the master table of tables:
        real_props = [
            (k, v)
            for (k, v) in self.jsonschema["properties"].items()
            if not k.startswith("$")
        ]
        for key, value in real_props:
            tables[master_table_name].append(
                dict(table=key, description=value.get("description", ""))
            )
        # then we get each table
        for key, value in real_props:
            tables[key] = self._get_table(value)
            # then we get column types
        example_inv = {1: "integer", "string": "string"}
        for key, value in real_props:
            rows = []
            if len(tables[key]) > 1:
                rows = [
                    dict(table=key, column=v, type="string") for v in ["key", "value"]
                ]
            if len(tables[key]) == 1:
                row1 = tables[key][0]
                rows = [
                    dict(table=key, column=k, type=example_inv[v])
                    for k, v in row1.items()
                ]
            tables[type_table_name].extend(rows)
        return tables

    @staticmethod
    def _get_table(contents):
        example = dict(integer=1, string="string")
        # several cases here:
        if contents["type"] == "object":
            # two columns: key-value in two columns
            properties = contents["properties"]
            return [
                dict(key=k, value=example[v["type"]]) for k, v in properties.items()
            ]
        # we're here, we're probably in an array
        assert contents["type"] == "array"
        items = contents["items"]
        if items["type"] != "object":
            # only one column with name
            return [example[items["type"]]]
        # here is a regular table:
        props = items["properties"]
        # if there are array of single values, we flatten them into one column:
        p_arrays = {
            k: v["items"]
            for k, v in props.items()
            if v["type"] == "array" and v["items"]["type"] != "object"
        }
        # if a column is an array of objects: we flatten the object into several columns
        p_arrays_objects = {
            "{}.{}".format(k, kk): vv["items"]
            for k, v in props.items()
            if v["type"] == "array" and v["items"]["type"] == "object"
            for kk, vv in v["items"]["properties"].items()
        }
        # the rest of columns stay the same
        p_no_array = {k: v for k, v in props.items() if v["type"] != "array"}
        props = {**p_arrays, **p_no_array, **p_arrays_objects}
        required = items["required"]
        rm_keys = props.keys() - set(required)
        # order is: first required in order, then the rest:
        one_line = {k: example[props[k]["type"]] for k in required}
        for k in rm_keys:
            one_line[k] = example[props[k]["type"]]
        return [one_line]

    @staticmethod
    def load_json(path):
        """
        Load a json file

        :param path: the path of the json file.json

        return the json content.
        """
        return load_json(path)

    @staticmethod
    def save_json(data, path):
        return save_json(data, path)

    """
    Aliases:
    """
    dict_to_flask = to_marshmallow
    load_schema = from_filepath
    jsonschema_to_flask = to_marshmallow
    jsonschema_to_dict = to_dict_schema



"""

"""

# Full imports
import json
import requests

# Partial imports
from marshmallow import ValidationError
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError, HTTPError

# Imports from modules
from cornflow_client import SchemaManager
from cornflow_client.constants import AirflowError, InvalidUsage


class Airflow(object):
    def __init__(self, url, user, pwd):
        self.url = url + "/api/v1"
        self.auth = HTTPBasicAuth(user, pwd)

    @classmethod
    def from_config(cls, config):
        data = dict(
            url=config["AIRFLOW_URL"],
            user=config["AIRFLOW_USER"],
            pwd=config["AIRFLOW_PWD"],
        )
        return cls(**data)

    def is_alive(self):
        try:
            response = requests.get(self.url + "/health")
        except (ConnectionError, HTTPError):
            return False
        try:
            data = response.json()
        except json.JSONDecodeError:
            return False
        return (
            data["metadatabase"]["status"] == "healthy"
            and data["scheduler"]["status"] == "healthy"
        )

    def request_headers_auth(self, status=200, **kwargs):
        def_headers = {"Content-type": "application/json", "Accept": "application/json"}
        headers = kwargs.get("headers", def_headers)
        response = requests.request(headers=headers, auth=self.auth, **kwargs)
        if status is None:
            return response
        if response.status_code != status:
            raise AirflowError(error=response.text, status_code=response.status_code)
        return response

    def consume_dag_run(self, dag_name, payload, dag_run_id=None, method="POST"):
        url = "{}/dags/{}/dagRuns".format(self.url, dag_name)
        if dag_run_id is not None:
            url = url + "/{}".format(dag_run_id)
        response = self.request_headers_auth(method=method, url=url, json=payload)
        return response

    def set_dag_run_state(self, dag_name, payload):
        url = "{}/dags/{}/updateTaskInstancesState".format(self.url, dag_name)
        return self.request_headers_auth(method="POST", url=url, json=payload)

    def run_dag(self, execution_id, dag_name="solve_model_dag"):
        conf = dict(exec_id=execution_id)
        payload = dict(conf=conf)
        return self.consume_dag_run(dag_name, payload=payload, method="POST")

    def update_schemas(self, dag_name="update_all_schemas"):
        return self.consume_dag_run(dag_name, payload={}, method="POST")

    def update_dag_registry(self, dag_name="update_dag_registry"):
        return self.consume_dag_run(dag_name, payload={}, method="POST")

    def get_dag_run_status(self, dag_name, dag_run_id):
        return self.consume_dag_run(
            dag_name, payload=None, dag_run_id=dag_run_id, method="GET"
        )

    def set_dag_run_to_fail(self, dag_name, dag_run_id, new_status="failed"):
        # here, two calls have to be done:
        # first we get information on the dag_run
        dag_run = self.consume_dag_run(
            dag_name, payload=None, dag_run_id=dag_run_id, method="GET"
        )
        dag_run_data = dag_run.json()
        # then, we use the "executed_date" to build a call to the change state api
        # TODO: We assume the solving task is named as is parent dag!
        payload = dict(
            dry_run=False,
            include_downstream=True,
            include_future=False,
            include_past=False,
            include_upstream=True,
            new_state=new_status,
            task_id=dag_name,
            execution_date=dag_run_data["execution_date"],
        )
        return self.set_dag_run_state(dag_name, payload=payload)

    def get_all_dag_runs(self, dag_name):
        return self.consume_dag_run(dag_name=dag_name, payload=None, method="GET")

    def get_dag_info(self, dag_name, method="GET"):
        url = "{}/dags/{}".format(self.url, dag_name)
        return self.request_headers_auth(method=method, url=url)

    def get_one_variable(self, variable):
        url = "{}/variables/{}".format(self.url, variable)
        return self.request_headers_auth(method="GET", url=url).json()

    def get_all_variables(self):
        return self.request_headers_auth(
            method="GET", url="{}/variables".format(self.url)
        ).json()

    def get_one_schema(self, dag_name, schema):
        return self.get_schemas_for_dag_name(dag_name)[schema]

    def get_schemas_for_dag_name(self, dag_name):
        response = self.get_one_variable(dag_name)
        result = json.loads(response["value"])
        result["name"] = response["key"]
        return result

    def get_all_schemas(self):
        response = self.get_all_variables()
        return [dict(name=variable["key"]) for variable in response["variables"]]

    def get_all_dags(self, method="GET"):
        url = "{}/dags".format(self.url)
        return self.request_headers_auth(method=method, url=url)

    def get_internal_dags(self, method="GET"):
        url = "{}/dags?tags=internal".format(self.url)
        return self.request_headers_auth(method=method, url=url)

    def get_model_dags(self, method="GET"):
        url = "{}/dags?tags=model".format(self.url)
        return self.request_headers_auth(method=method, url=url)


def get_schema(config, dag_name, schema="instance"):
    """
    Gets a schema by name from airflow server. We use the variable api.
    We transform the jsonschema into a marshmallow class

    """
    af_client = Airflow.from_config(config)
    if not af_client.is_alive():
        raise AirflowError(error="Airflow is not accessible")

    schema_json = af_client.get_one_schema(dag_name, schema)
    manager = SchemaManager(schema_json)
    return manager.jsonschema_to_flask()


def validate_and_continue(obj, data):
    """
    This function validates some data meets a marshmallow object specs.
    In case it does not: we raise an error.
    In case we do, we return the transformed data

    """
    try:
        validate = obj.load(data)
    except ValidationError as e:
        raise InvalidUsage(error="Bad data format: {}".format(e))
    err = ""
    if validate is None:
        raise InvalidUsage(error="Bad data format: {}".format(err))
    return validate



"""

"""
# Partial imports
from abc import ABC

# Imports from internal modules
from .instance_solution import InstanceSolutionCore


class SolutionCore(InstanceSolutionCore, ABC):
    """
    The solution template.
    """

    pass



"""

"""
import json
import os
import time
from unittest import TestCase

from cornflow_client import CornFlow
from cornflow_client.airflow.api import Airflow
from cornflow_client.constants import STATUS_OPTIMAL, STATUS_NOT_SOLVED
from cornflow_client.tests.const import PULP_EXAMPLE

# Constants
path_to_tests_dir = os.path.dirname(os.path.abspath(__file__))

# Helper functions
def _load_file(_file):
    with open(_file) as f:
        temp = json.load(f)
    return temp


def _get_file(relative_path):
    return os.path.join(path_to_tests_dir, relative_path)


class TestAirflowClient(TestCase):
    def setUp(self):
        self.client = Airflow(url="http://127.0.0.1:8080", user="admin", pwd="admin")

    def test_alive(self):
        self.assertTrue(self.client.is_alive())

    def test_connect_from_config(self):
        client = Airflow.from_config(
            {
                "AIRFLOW_URL": "http://127.0.0.1:8080",
                "AIRFLOW_USER": "admin",
                "AIRFLOW_PWD": "admin",
            }
        )
        self.assertTrue(client.is_alive())

    def test_bad_connection(self):
        client = Airflow(url="http://127.0.0.1:8088", user="admin", pwd="admin!")
        self.assertFalse(client.is_alive())

    def test_update_schemas(self):
        response = self.client.update_schemas()
        self.assertEqual(200, response.status_code)

    def test_update_dag_registry(self):
        response = self.client.update_dag_registry()
        self.assertEqual(200, response.status_code)

    def test_run_dag(self):
        data = _load_file(PULP_EXAMPLE)
        cf_client = CornFlow(url="http://127.0.0.1:5050/")
        cf_login = cf_client.login("user", "UserPassword1!")
        instance = cf_client.create_instance(data, "test_example", "test_description")
        execution = cf_client.create_execution(
            instance_id=instance["id"],
            config={"solver": "PULP_CBC_CMD", "timeLimit": 100},
            name="test_execution",
            description="execution_description",
            schema="solve_model_dag",
            run=False,
        )

        # Check that execution is not run
        status = cf_client.get_status(execution_id=execution["id"])
        self.assertEqual(-4, status["state"])

        # Run the execution
        response = self.client.run_dag(execution_id=execution["id"])
        self.assertEqual(200, response.status_code)
        self.assertIn("dag_run_id", response.json().keys())

        # Check that is optimal
        time.sleep(10)
        status = cf_client.get_status(execution_id=execution["id"])
        self.assertEqual(1, status["state"])



from cornflow_client import (
    InstanceCore,
    SolutionCore,
    ApplicationCore,
    ExperimentCore,
    get_empty_schema,
)
from cornflow_client.constants import BadConfiguration, BadInstance
import unittest


class TestABC(unittest.TestCase):
    def test_good_solution(self):
        GoodSolutionClass(dict(a=1))

    def test_bad_solution(self):
        must_fail = lambda: BadSolutionClass(dict(a=1))
        self.assertRaises(TypeError, must_fail)

    def test_good_instance(self):
        GoodInstanceClass(dict(a=1))

    def test_bad_instance(self):
        must_fail = lambda: BadInstanceClass(dict(a=1))
        self.assertRaises(TypeError, must_fail)

    def test_experiment(self):
        GoodExperiment(GoodInstanceClass(dict()), GoodSolutionClass(dict()))

    def test_bad_experiment(self):
        must_fail = lambda: BadExperiment(
            GoodInstanceClass(dict()), GoodSolutionClass(dict())
        )
        self.assertRaises(TypeError, must_fail)

    def test_good_application(self):
        GoodApp().get_solver("default")

    def test_good_application_solver(self):
        solver = ConcatenatedSolver().get_solver("pulp")
        self.assertIsNotNone(solver)

    def test_good_application_solver_none(self):
        solver = ConcatenatedSolver().get_solver("pulp1")
        self.assertIsNone(solver)

    def test_bad_application(self):
        must_fail = lambda: BadApp()
        self.assertRaises(TypeError, must_fail)

    def test_bad_configuration(self):
        must_fail = lambda: GoodApp().solve(
            data=dict(number=1), config=dict(timeLimit="")
        )
        self.assertRaises(BadConfiguration, must_fail)

    def test_incorrect_good_instance(self):
        must_fail = lambda: GoodApp().solve(data=dict(number=""), config=dict())
        self.assertRaises(BadInstance, must_fail)


class GoodInstanceClass(InstanceCore):
    schema = get_empty_schema(dict(number=dict(type="number")))


class BadInstanceClass(InstanceCore):
    @classmethod
    def from_dict(cls, data: dict) -> "BadInstanceClass":
        return cls(data)


class GoodSolutionClass(SolutionCore):
    def schema(self):
        return dict()


class BadSolutionClass(SolutionCore):
    @classmethod
    def from_dict(cls, data: dict) -> "BadSolutionClass":
        return cls(data)


class GoodExperiment(ExperimentCore):
    def solve(self, options: dict):
        return dict(sol_status=1, status=1)

    def get_objective(self) -> float:
        return 1

    def check_solution(self, *args, **kwargs) -> dict:
        return dict()


class BadExperiment(ExperimentCore):
    def solve(self, options) -> dict:
        return dict()


class GoodApp(ApplicationCore):
    name = "123"
    instance = GoodInstanceClass
    solution = GoodSolutionClass
    solvers = dict(default=GoodExperiment)
    schema = get_empty_schema(dict(timeLimit=dict(type="number")), solvers=["default"])
    test_cases = [dict()]


class ConcatenatedSolver(ApplicationCore):
    name = "123"
    instance = GoodInstanceClass
    solution = GoodSolutionClass
    solvers = dict(pulp=GoodExperiment)
    schema = get_empty_schema(dict(timeLimit=dict(type="number")), solvers=["pulp.cbc"])
    test_cases = [dict()]


class BadApp(ApplicationCore):
    name = "123"
    solvers = dict(default=GoodExperiment)



"""

"""
# Full imports
import json
import pickle

# Partial imports
from pytups import OrderSet


def new_set(seq):
    """
    :param seq: a (hopefully unique) list of elements (tuples, strings, etc.)
    Returns a new ordered set
    """
    return OrderSet(seq)


def load_json(path):
    with open(path) as json_file:
        file = json.load(json_file)
    return file


def save_json(data, path):
    with open(path, "w") as outfile:
        json.dump(data, outfile)


def copy(dictionary):
    return pickle.loads(pickle.dumps(dictionary, -1))



"""

"""
#
import json
import os

#
from unittest import TestCase

#
from cornflow_client import SchemaManager
from cornflow_client.constants import DATASCHEMA
from cornflow_client.core.tools import load_json
from cornflow_client.tests.const import dict_example


class TestSchemaManager(TestCase):
    def setUp(self):
        self.root_data = os.path.join(os.path.dirname(__file__), "../data")
        pass

    def get_data_file(self, filename):
        return os.path.join(self.root_data, filename)

    def get_project_data_file(self, filename):
        return os.path.join(self.root_data, "../../data", filename)

    def test_schema_dict(self):
        sm = SchemaManager.from_filepath(
            self.get_project_data_file("../data/pulp_json_schema.json")
        )
        dict_schema = sm.jsonschema_to_dict()

        self.assertCountEqual(
            dict_schema["CoefficientSchema"], dict_example["CoefficientSchema"]
        )
        self.assertCountEqual(
            dict_schema["ObjectiveSchema"], dict_example["ObjectiveSchema"]
        )
        self.assertCountEqual(
            dict_schema["ConstraintsSchema"], dict_example["ConstraintsSchema"]
        )
        self.assertCountEqual(
            dict_schema["VariablesSchema"], dict_example["VariablesSchema"]
        )
        self.assertCountEqual(
            dict_schema["ParametersSchema"], dict_example["ParametersSchema"]
        )
        self.assertCountEqual(dict_schema["Sos1Schema"], dict_example["Sos1Schema"])
        self.assertCountEqual(dict_schema["Sos2Schema"], dict_example["Sos2Schema"])
        self.assertCountEqual(dict_schema[DATASCHEMA], dict_example[DATASCHEMA])
        sm.jsonschema_to_flask()

    def test_schema_validation(self):
        sm = SchemaManager.from_filepath(
            self.get_project_data_file("pulp_json_schema.json")
        )
        val = sm.validate_file(self.get_data_file("pulp_example_data.json"))
        self.assertTrue(val)
        sm.jsonschema_to_flask()

    def test_schema_validation_2(self):
        sm = SchemaManager.from_filepath(self.get_data_file("hk_data_schema.json"))
        val = sm.validate_file(self.get_data_file("hk_data_input.json"))
        self.assertTrue(val)
        # Test that it can be transformed into a dict
        dict_schema = sm.jsonschema_to_dict()
        self.assertEqual(
            dict_schema["JobsSchema"][0],
            {
                "name": "id",
                "type": "Integer",
                "required": True,
                "allow_none": False,
                "many": False,
            },
        )
        self.assertEqual(
            dict_schema["JobsSchema"][1],
            {"name": "successors", "type": "Integer", "many": True, "required": True},
        )
        marshmallow_object = sm.dict_to_flask()
        self.assertEqual(
            marshmallow_object().fields.keys(),
            {"resources", "needs", "jobs", "durations"},
        )
        with open(self.get_data_file("hk_data_input.json"), "r") as f:
            content = json.load(f)
        marshmallow_object().load(content)
        # marshmallow_object().fields['jobs'].nested().fields['successors']

    def test_check_schema(self):
        sm = SchemaManager.from_filepath(
            self.get_project_data_file("vrp_solution_schema.json")
        )
        val = sm.validate_schema()
        self.assertTrue(val)
        sm.jsonschema_to_flask()

    def test_check_wrong_schema_1(self):
        schema = load_json(self.get_project_data_file("vrp_solution_schema.json"))
        del schema["properties"]["routes"]["items"]["required"]
        sm = SchemaManager(schema)
        val = sm.validate_schema()
        self.assertFalse(val)

    def test_check_wrong_schema_2(self):
        schema = load_json(self.get_project_data_file("vrp_solution_schema.json"))
        schema["properties"]["routes"]["items"]["properties"]["pos"][
            "type"
        ] = "not_a_type"
        sm = SchemaManager(schema)
        val = sm.validate_schema()
        self.assertFalse(val)

    def test_check_wrong_schema_3(self):
        schema = load_json(self.get_project_data_file("vrp_solution_schema.json"))
        del schema["properties"]["routes"]["items"]["properties"]["pos"]["type"]
        sm = SchemaManager(schema)
        val = sm.validate_schema()
        self.assertFalse(val)

    def test_validation_errors(self):
        sm = SchemaManager.from_filepath(
            self.get_project_data_file("pulp_json_schema.json")
        )
        data = {"objective": [], "constraints": [], "variables": []}
        bool = sm.validate_data(data)
        val = sm.get_validation_errors(data)
        self.assertFalse(bool)
        # The two errors are: objective should be an object and parameters is missing.
        # we do not impose sos1 and sos2 to be there.
        self.assertEqual(len(val), 2)
        self.assertEqual(val[0].message, "[] is not of type 'object'")
        sm.jsonschema_to_flask()

    def test_validation_errors2(self):
        sm = SchemaManager.from_filepath(
            self.get_project_data_file("pulp_json_schema.json")
        )
        data = {
            "objective": [],
            "constraints": ["notAConstraint"],
            "variables": ["notAVariable"],
        }
        val = sm.get_validation_errors(data)
        # errors are: objective should be an object, there's a wrongly formatted constraint and variable.
        # and parameters is missing
        self.assertEqual(len(val), 4)
        sm.jsonschema_to_flask()

    def test_validation_errors3(self):
        sm = SchemaManager.from_filepath(self.get_data_file("hk_data_schema.json"))
        bool = sm.validate_file(self.get_data_file("data_input_bad.json"))
        val = sm.get_file_errors(self.get_data_file("data_input_bad.json"))
        self.assertFalse(bool)
        self.assertEqual(len(val), 2)
        sm.jsonschema_to_flask()

    def test_schema_names(self):
        sm = SchemaManager.from_filepath(self.get_data_file("name_problem_schema.json"))
        dict_schema = sm.jsonschema_to_dict()
        self.assertEqual(len(dict_schema["CoefficientsSchema"]), 2)
        self.assertEqual(len(dict_schema["Coefficients1Schema"]), 1)
        sm.jsonschema_to_flask()

    def test_array_integer(self):
        sm = SchemaManager.from_filepath(
            self.get_data_file("graph_coloring_input.json")
        )
        sm.jsonschema_to_flask()

    def test_non_mandatory(self):
        sm = SchemaManager.from_filepath(self.get_data_file("instance-hackathon2.json"))
        schema_marsh = sm.jsonschema_to_flask()
        with open(self.get_data_file("hk_data_input.json"), "r") as f:
            content = json.load(f)
        err = schema_marsh().load(content)
        return

    def test_flask_schema_extra_info(self):
        with open(self.get_data_file("pulp_example_data.json"), "r") as f:
            content = json.load(f)
        sm = SchemaManager.from_filepath(
            self.get_project_data_file("pulp_json_schema.json")
        )
        marshmallow_object = sm.jsonschema_to_flask()
        content["new_param"] = 1
        content["objective"]["another_something_new"] = 1
        marshmallow_object().load(content)

    def test_to_excel(self):
        sm = SchemaManager.from_filepath(
            self.get_data_file("graph_coloring_input.json")
        )
        template = sm.to_template()
        import pandas as pd

        dataframes = {k: pd.DataFrame.from_dict(v) for k, v in template.items()}
        self.assertEqual(dataframes["pairs"]["n1"][0], 1)
        self.assertEqual(dataframes["pairs"].shape, (1, 2))
        self.assertEqual(dataframes["_README"].shape, (1, 2))
        self.assertEqual(dataframes["_TYPES"].shape, (2, 3))

    def test_to_excel2(self):
        sm = SchemaManager.from_filepath(self.get_data_file("hk_data_schema.json"))
        template = sm.to_template()
        import pandas as pd

        dataframes = {k: pd.DataFrame.from_dict(v) for k, v in template.items()}
        self.assertEqual(dataframes["durations"].job[0], 1)
        self.assertEqual(dataframes["durations"].shape, (1, 3))
        self.assertEqual(dataframes["jobs"].shape, (1, 2))
        self.assertEqual(dataframes["_README"].shape, (4, 2))
        self.assertEqual(dataframes["_TYPES"].shape, (11, 3))

    def test_to_excel3(self):
        path = self.get_data_file("../../data/empty_schema.json")
        schema = SchemaManager.load_json(path)
        schema["properties"] = dict(
            configuration=dict(
                type="object",
                properties=dict(
                    option1=dict(type="integer"), option2=dict(type="string")
                ),
            )
        )
        sm = SchemaManager(schema)
        template = sm.to_template()
        import pandas as pd

        dataframes = {k: pd.DataFrame.from_dict(v) for k, v in template.items()}
        self.assertEqual(len(dataframes["configuration"]), 2)
        self.assertEqual(dataframes["configuration"].iloc[1, 1], "string")
        self.assertEqual(dataframes["configuration"].iloc[0, 1], 1)
        self.assertEqual(dataframes["_README"].shape, (1, 2))
        self.assertEqual(dataframes["_TYPES"].shape, (2, 3))

    # TODO: fix this test and uncomment
    # def test_list_of_lists(self):
    #     sm = SchemaManager.from_filepath(self.get_data_file('graph_coloring2_input.json'))
    #     sm.jsonschema_to_flask()



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

required = []
with open("requirements.txt", "r") as fh:
    required.append(fh.read().splitlines())

extra_required = {"excel": ["openpyxl", "pandas"]}


setuptools.setup(
    name="cornflow-client",
    version="0.36.0",
    author="baobab soluciones",
    author_email="sistemas@baobabsoluciones.es",
    description="Client to connect to a cornflow server",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/baobabsoluciones/cornflow-client",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
    ],
    python_requires=">=3.7",
    include_package_data=True,
    install_requires=required,
    extra_require=extra_required,
)



"""

"""
# Partial imports
from abc import ABC

# Imports from internal modules
from .instance_solution import InstanceSolutionCore


class InstanceCore(InstanceSolutionCore, ABC):
    """
    The instance template.
    """

    # TODO: make abstractmethod
    def check_inconsistencies(self, *args, **kwargs) -> dict:
        """
        Method that checks if there are inconsistencies in the data of the current instance.

        :return: A dictionary containing the inconsistencies found.
        """
        return dict()

    # TODO: make abstractmethod
    def check_feasibility(self, *args, **kwargs) -> bool:
        """
        Method that checks if the problem is feasible.

        :return: True if the problem is feasible, False otherwise
        """
        return True

    # TODO: make abstractmethod
    def check(self, *args, **kwargs) -> dict:
        """
        Method that checks if there are inconsistencies in the data of the instance and if the problem is feasible

        :return: An dictionary containing the inconsistencies found and indicating if the problem is infeasible
        """
        inconsistencies = self.check_inconsistencies(*args, **kwargs)
        is_feasible = self.check_feasibility(*args, **kwargs)
        if not is_feasible:
            inconsistencies["is_infeasible"] = True
        return inconsistencies



"""

"""
# Full imports
import logging as log
import re
import requests

# Partial imports
from functools import wraps
from urllib.parse import urljoin


class CornFlow(object):
    """
    Base class to access cornflow-server
    """

    def __init__(self, url, token=None):
        self.url = url
        self.token = token

    def ask_token(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.token:
                raise CornFlowApiError("Need to login first!")
            return func(self, *args, **kwargs)

        return wrapper

    def log_call(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            log.debug(result)
            return result

        return wrapper

    def prepare_encoding(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            encoding = kwargs.get("encoding", "br")
            if encoding not in [
                "gzip",
                "compress",
                "deflate",
                "br",
                "identity",
            ]:
                encoding = "br"
            kwargs["encoding"] = encoding
            result = func(*args, **kwargs)
            return result

        return wrapper

    # def expect_201(func):
    #     return partial(expect_status, status=201)
    #
    # def expect_200(func):
    #     return partial(expect_status, status=200)

    def api_for_id(self, api, id, method, post_url="", encoding=None, **kwargs):
        """
        :param api: the resource in the server
        :param id: the id of the particular object
        :param method: HTTP method to apply
        :param post_url: optional action to apply
        :param encoding: optional string with the type of encoding, if it is not specified it uses br encoding,
        options are: gzip, compress, deflate, br or identity
        :param kwargs: other arguments to requests.request

        :return: requests.request
        """
        if post_url and post_url[-1] != "/":
            post_url += "/"
        url = urljoin(urljoin(self.url, api) + "/", str(id) + "/" + post_url)
        return requests.request(
            method=method,
            url=url,
            headers={
                "Authorization": "access_token " + self.token,
                "Content-Encoding": encoding,
            },
            **kwargs,
        )

    def get_api(self, api, method="GET", encoding=None, **kwargs):
        return requests.request(
            method=method,
            url=urljoin(self.url, api) + "/",
            headers={
                "Authorization": "access_token " + self.token,
                "Content-Encoding": encoding,
            },
            **kwargs,
        )

    @ask_token
    @prepare_encoding
    def get_api_for_id(self, api, id, post_url="", encoding=None, **kwargs):
        """
        api_for_id with a GET request
        """

        return self.api_for_id(
            api=api, id=id, post_url=post_url, encoding=encoding, **kwargs, method="get"
        )

    @ask_token
    @prepare_encoding
    def delete_api_for_id(self, api, id, encoding=None, **kwargs):
        """
        api_for_id with a DELETE request
        """
        return self.api_for_id(
            api=api, id=id, encoding=encoding, **kwargs, method="delete"
        )

    @ask_token
    @prepare_encoding
    def put_api_for_id(self, api, id, payload, encoding=None, **kwargs):
        """
        api_for_id with a PUT request
        """
        return self.api_for_id(
            api=api, id=id, json=payload, method="put", encoding=encoding, **kwargs
        )

    @ask_token
    @prepare_encoding
    def patch_api_for_id(self, api, id, payload, encoding=None, **kwargs):
        """
        api_for_id with a PATCH request
        """
        return self.api_for_id(
            api=api, id=id, json=payload, method="patch", encoding=encoding, **kwargs
        )

    @ask_token
    @prepare_encoding
    def post_api_for_id(self, api, id, encoding=None, **kwargs):
        """
        api_for_id with a POST request
        """
        return self.api_for_id(
            api=api, id=id, method="post", encoding=encoding, **kwargs
        )

    @ask_token
    @prepare_encoding
    def create_api(self, api, encoding=None, **kwargs):
        return requests.post(
            urljoin(self.url, api),
            headers={
                "Authorization": "access_token " + self.token,
                "Content-Encoding": encoding,
            },
            **kwargs,
        )

    @log_call
    @prepare_encoding
    def sign_up(self, username, email, pwd, encoding=None):
        """
        Sign-up to the server. Creates a new user or returns error.

        :param str username: username
        :param str email: email
        :param str pwd: password
        :param str encoding: the type of encoding used in the call. Defaults to 'br'

        :return: a dictionary with a token inside
        """
        return requests.post(
            urljoin(self.url, "signup/"),
            json={"username": username, "password": pwd, "email": email},
            headers={"Content-Encoding": encoding},
        )

    @log_call
    def is_alive(self):
        """
        Asks the server if it's alive
        """
        response = requests.get(urljoin(self.url, "health/"))
        if response.status_code == 200:
            return response.json()
        raise CornFlowApiError(
            "Connection failed with status code: {}: {}".format(
                response.status_code, response.text
            )
        )

    @prepare_encoding
    def login(self, username, pwd, encoding=None):
        """
        Log-in to the server.

        :param str username: username
        :param str pwd: password
        :param str encoding: the type of encoding used in the call. Defaults to 'br'

        :return: a dictionary with a token inside
        """
        response = requests.post(
            urljoin(self.url, "login/"),
            json={"username": username, "password": pwd},
            headers={"Content-Encoding": encoding},
        )
        if response.status_code == 200:
            result = response.json()
            self.token = result["token"]
            return result
        else:
            raise CornFlowApiError(
                "Login failed with status code: {}: {}".format(
                    response.status_code, response.text
                )
            )

    # TODO: those status_code checks should be done via a decorator. But I do not know how.
    @ask_token
    @log_call
    @prepare_encoding
    def create_instance(
        self, data, name=None, description="", schema="solve_model_dag", encoding=None
    ):
        """
        Uploads an instance to the server

        :param dict data: data according to json-schema for the problem
        :param str name: name for instance
        :param str description: description of the instance
        :param str schema: name of problem to solve
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        if name is None:
            try:
                name = data["parameters"]["name"]
            except IndexError:
                raise CornFlowApiError("The `name` argument needs to be filled")
        payload = dict(data=data, name=name, description=description, schema=schema)
        response = self.create_api("instance/", json=payload, encoding=encoding)
        if response.status_code != 201:
            raise CornFlowApiError(
                "Expected a code 201, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()

    @ask_token
    @log_call
    @prepare_encoding
    def create_case(
        self,
        name,
        schema,
        data=None,
        parent_id=None,
        description="",
        solution=None,
        encoding=None,
    ):
        """
        Uploads a case to the server

        :param dict data: data according to json-schema for the problem
        :param str name: name for case
        :param str description: description of case
        :param str schema: name of problem of case
        :param str parent_id: id of the parent directory in the tree structure
        :param dict solution: optional solution data to store inside case
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        payload = dict(
            name=name,
            description=description,
            schema=schema,
            parent_id=parent_id,
        )
        # Both data AND solution are optional.
        # data is optional in directories.
        # solutions in unsolved instances
        if data is not None:
            payload["data"] = data
        if solution is not None:
            payload["solution"] = solution
        response = self.create_api("case/", json=payload, encoding=encoding)
        if response.status_code != 201:
            raise CornFlowApiError(
                "Expected a code 201, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()

    @ask_token
    @log_call
    @prepare_encoding
    def create_instance_file(
        self, filename, name, description="", minimize=True, encoding=None
    ):
        """
        Uploads a file to the server to be parsed into an instance

        :param str filename: path to filename to upload
        :param str name: name for instance
        :param str description: description of the instance
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        with open(filename, "rb") as file:
            response = self.create_api(
                "instancefile/",
                data=dict(name=name, description=description, minimize=minimize),
                files=dict(file=file),
                encoding=encoding,
            )

        if response.status_code != 201:
            raise CornFlowApiError(
                "Expected a code 201, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def create_execution(
        self,
        instance_id,
        config,
        name="test1",
        description="",
        schema="solve_model_dag",
        encoding=None,
        run=True,
    ):
        """
        Creates an execution from a (previously) uploaded instance

        :param str instance_id: id for the instance
        :param str name: name for the execution
        :param str description: description of the execution
        :param dict config: execution configuration
        :param str schema: name of the problem to solve
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        :param bool run: if the execution should be run or not
        """
        api = "execution/"
        payload = dict(
            config=config,
            instance_id=instance_id,
            name=name,
            description=description,
            schema=schema,
        )
        if not run:
            api += "?run=0"
        response = self.create_api(api, json=payload, encoding=encoding)
        if response.status_code != 201:
            raise CornFlowApiError(
                "Expected a code 201, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()

    @ask_token
    @prepare_encoding
    def get_data(self, execution_id, encoding=None):
        """
        Downloads the data from an execution

        :param str execution_id: id for the execution
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.get_api_for_id(api="dag/", id=execution_id, encoding=encoding)
        return response.json()

    @ask_token
    @prepare_encoding
    def write_solution(self, execution_id, encoding=None, **kwargs):
        """
        Edits an execution

        :param str execution_id: id for the execution
        :param kwargs: optional data to edit
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.put_api_for_id(
            "dag/", id=execution_id, encoding=encoding, payload=kwargs
        )
        if response.status_code != 200:
            raise CornFlowApiError(
                "Expected a code 200, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()

    @ask_token
    @prepare_encoding
    def write_instance_checks(self, instance_id, encoding=None, **kwargs):
        """"""
        response = self.put_api_for_id(
            "dag/instance/", id=instance_id, encoding=encoding, payload=kwargs
        )
        if response.status_code != 200:
            raise CornFlowApiError(
                f"Expected a 200, got a {response.status_code} error instead: {response.text}"
            )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def stop_execution(self, execution_id, encoding=None):
        """
        Interrupts an ongoing execution

        :param str execution_id: id for the execution
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.post_api_for_id(
            api="execution/", id=execution_id, encoding=encoding
        )
        if response.status_code != 200:
            raise CornFlowApiError(
                "Expected a code 200, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()

    @ask_token
    @prepare_encoding
    def manual_execution(self, instance_id, config, name, encoding=None, **kwargs):
        """
        Uploads an execution from solution data

        :param str instance_id: id for the instance
        :param str name: name for execution
        :param str description: description of the execution
        :param dict config: execution configuration
        :param str schema: name of the problem to solve
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        :param kwargs: contents of the solution (inside data)
        """
        payload = dict(config=config, instance_id=instance_id, name=name, **kwargs)
        response = self.create_api("dag/", json=payload, encoding=encoding)
        if response.status_code != 201:
            raise CornFlowApiError(
                "Expected a code 201, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_results(self, execution_id, encoding=None):
        """
        Downloads results from an execution

        :param str execution_id: id for the execution
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.get_api_for_id(
            api="execution/", id=execution_id, encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_status(self, execution_id, encoding=None):
        """
        Downloads the current status of an execution

        :param str execution_id: id for the execution
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.get_api_for_id(
            api="execution/", id=execution_id, post_url="status", encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_log(self, execution_id, encoding=None):
        """
        Downloads the log for an execution

        :param str execution_id: id for the execution
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.get_api_for_id(
            api="execution/", id=execution_id, post_url="log", encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_solution(self, execution_id, encoding=None):
        """
        Downloads the solution data for an execution

        :param str execution_id: id for the execution
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.get_api_for_id(
            api="execution/", id=execution_id, post_url="data", encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_all_instances(self, params=None, encoding=None):
        """
        Downloads all the user's instances

        :param dict params: optional filters
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        return self.get_api("instance", params=params, encoding=encoding).json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_all_cases(self, params=None, encoding=None):
        """
        Downloads all the user's cases

        :param dict params: optional filters
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        return self.get_api("case", params=params, encoding=encoding).json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_all_executions(self, params=None, encoding=None):
        """
        Downloads all the user's executions

        :param dict params: optional filters
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        return self.get_api("execution", params=params, encoding=encoding).json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_all_users(self, encoding=None):
        """
        Downloads all the users in the server
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        return self.get_api("user", encoding=encoding).json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_one_user(self, user_id, encoding=None):
        """
        Downloads all information on a user

        :param str user_id: user id
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        return self.get_api_for_id(api="user", id=user_id, encoding=encoding)

    @log_call
    @ask_token
    @prepare_encoding
    def get_one_instance(self, reference_id, encoding=None):
        """
        Downloads header of an instance

        :param str reference_id: id for the instance
        :param str encoding: the type of encoding used in the call. Defaults to 'br'

        """
        response = self.get_api_for_id(
            api="instance", id=reference_id, encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_one_case(self, reference_id, encoding=None):
        """
        Downloads header of a case

        :param str reference_id: id for the case
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.get_api_for_id(api="case", id=reference_id, encoding=encoding)
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def delete_one_case(self, reference_id, encoding=None):
        """
        Deletes a case

        :param str reference_id: id for the case
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.delete_api_for_id(
            api="case", id=reference_id, encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def put_one_case(self, reference_id, payload, encoding=None):
        """
        Edits a case

        :param str reference_id: id for the case
        :param dict payload: data to edit
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.put_api_for_id(
            api="case", id=reference_id, payload=payload, encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def patch_one_case(self, reference_id, payload, encoding=None):
        """
        Patch a case

        :param str reference_id: id for the case
        :param dict payload: data to edit
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.patch_api_for_id(
            api="case", id=reference_id, payload=payload, encoding=encoding
        )
        return response.json()

    @log_call
    @ask_token
    @prepare_encoding
    def delete_one_instance(self, reference_id, encoding=None):
        """
        Delete an instance

        :param str reference_id: id for the instance
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.delete_api_for_id(
            api="instance", id=reference_id, encoding=encoding
        )
        return response

    @ask_token
    @prepare_encoding
    def get_schema(self, dag_name, encoding=None):
        """
        Downloads schemas for a problem

        :param str dag_name: id for the problem
        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        response = self.get_api_for_id(api="schema", id=dag_name, encoding=encoding)
        return response.json()

    @ask_token
    @prepare_encoding
    def get_all_schemas(self, encoding=None):
        """
        Downloads all problems' (aka as app's) names

        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        return self.get_api("schema", encoding=encoding).json()

    @log_call
    @ask_token
    @prepare_encoding
    def get_deployed_dags(self, encoding=None):
        """
        Downloads the deployed dags in cornflow in order to update them

        :param str encoding: the type of encoding used in the call. Defaults to 'br'
        """
        return self.get_api("dag/deployed", encoding=encoding).json()

    @log_call
    @ask_token
    @prepare_encoding
    def create_deployed_dag(
        self, name: str = None, description: str = None, encoding=None
    ):
        if name is None:
            return {"error": "No dag name was given"}
        payload = dict(id=name, description=description)
        response = self.create_api("dag/deployed/", json=payload, encoding=encoding)
        if response.status_code != 201:
            raise CornFlowApiError(
                "Expected a code 201, got a {} error instead: {}".format(
                    response.status_code, response.text
                )
            )
        return response.json()


class CornFlowApiError(Exception):
    """
    CornFlow returns an error
    """

    pass


def arg_to_value(
    some_string, replace_underscores_with_spaces=False, force_number=False
):
    if replace_underscores_with_spaces:
        some_string = re.sub(pattern=r"_", repl=r" ", string=some_string)
    if some_string[0] == "'" and some_string[-1] == "'":
        # this means its a string, no need to continue
        return some_string[1:-1]
    if not force_number:
        return some_string
    if some_string.isdigit():
        return int(some_string)
    # TODO: probably other edge cases, such as boolean.
    try:
        return float(some_string)
    except ValueError:
        return some_string


def get_tuple_from_root(rest: str, **kwargs):
    if not rest:
        # it matches exactly, no index.
        raise CornFlowApiError("There is not rest: there is just one variable")

    if rest[0] == "(":
        # key is a tuple.
        args = re.split(",_?", rest[1:-1])
        kwargs = {**kwargs, "force_number": True}
        new_args = [arg_to_value(arg, **kwargs) for arg in args if arg != ""]
        return tuple(new_args)
    # key is a single value
    return arg_to_value(rest, **kwargs)


def group_variables_by_name(_vars, names_list, **kwargs):
    # this is an experimental function that assumes the following:
    # 1. keys do not have special characters: -+[] ->/
    # 2. key can be a tuple or a single string.
    # 3. if a tuple, they can be an integer or a string.
    #
    # it dos not permit the nested dictionary format of variables
    # we copy it because we will be taking out already seen variables
    _vars = dict(_vars)
    __vars = {k: {} for k in names_list}
    for root in names_list:
        # 1. match root name to variables
        candidates = {name: obj for name, obj in _vars.items() if name.startswith(root)}
        # 2 extract and store said variables
        try:
            __vars[root] = {
                get_tuple_from_root(name[len(root) + 1 :], **kwargs): obj
                for name, obj in candidates.items()
            }
        except CornFlowApiError:
            __vars[root] = list(candidates.values())[0]
        # 3. take out from list
        for name in candidates:
            _vars.pop(name)
    return __vars



from .application import ApplicationCore
from .experiment import ExperimentCore
from .instance_solution import InstanceSolutionCore
from .instance import InstanceCore
from .solution import SolutionCore
from .extra_methods import *



from .cornflow_client import CornFlow, group_variables_by_name, CornFlowApiError
from cornflow_client.schema.manager import SchemaManager
from pytups import TupList, SuperDict, OrderSet
from cornflow_client.core import (
    ApplicationCore,
    InstanceCore,
    SolutionCore,
    ExperimentCore,
)
from cornflow_client.schema.tools import get_empty_schema, get_pulp_jsonschema



import unittest
import pulp
from cornflow_client import group_variables_by_name


class VarsGroup(unittest.TestCase):

    # TODO: something should be asserted
    def test_export_pb_vars_1(self):
        prob = pulp.LpProblem("test_export_pb_vars_1", pulp.LpMaximize)
        x = pulp.LpVariable("x", 0, 4)
        y = pulp.LpVariable("y", -1, 1)
        z = pulp.LpVariable("z", 0)
        w = pulp.LpVariable("w", 0)
        prob += x + 4 * y + 9 * z, "obj"
        prob += x + y <= 5, "c1"
        prob += x + z >= 10, "c2"
        prob += -y + z == 7, "c3"
        prob += w >= 0, "c4"
        data = prob.to_dict()
        var1, prob1 = pulp.LpProblem.from_dict(data)
        group_variables_by_name(var1, ["x", "y", "z", "w"])
        x, y, z, w = [var1[name] for name in ["x", "y", "z", "w"]]

    # TODO: something should be asserted
    def test_export_pb_vars_2(self):
        prob = pulp.LpProblem("test_export_pb_vars_2", pulp.LpMaximize)
        x = pulp.LpVariable("decision_x", 0, 4)
        y = pulp.LpVariable("decision_y", -1, 1)
        z = pulp.LpVariable("decision_z", 0)
        w = pulp.LpVariable("decision_special_w", 0)
        prob += x + 4 * y + 9 * z, "obj"
        prob += x + y <= 5, "c1"
        prob += x + z >= 10, "c2"
        prob += -y + z == 7, "c3"
        prob += w >= 0, "c4"
        data = prob.to_dict()
        var1, prob1 = pulp.LpProblem.from_dict(data)
        grouped_dict = group_variables_by_name(var1, ["decision_special", "decision"])
        x, y, z = [grouped_dict["decision"][name] for name in ["x", "y", "z"]]
        w = grouped_dict["decision_special"]["w"]

    def test_read_vars_stringTuples(self):
        plants = ["San_Francisco", "Los_Angeles", "Phoenix", "Denver"]
        stores = ["San_Diego", "Barstow", "Tucson", "Dallas"]

        plant_store = [(p, s) for p in plants for s in stores]
        # Creates the problem variables of the Flow on the Arcs
        flow = pulp.LpVariable.dicts("Route", plant_store, 0, None, pulp.LpInteger)
        _flow = {v.name: v for v in flow.values()}
        a = group_variables_by_name(_flow, ["Route"])
        self.assertEqual(flow, a["Route"])

    def test_read_vars_intStringTuples(self):
        plants = range(1, 5)
        stores = ["San_Diego", "Barstow", "Tucson", "Dallas"]

        plant_store = [(p, s) for p in plants for s in stores]
        # Creates the problem variables of the Flow on the Arcs
        flow = pulp.LpVariable.dicts("Route", plant_store, 0, None, pulp.LpInteger)
        _flow = {v.name: v for v in flow.values()}
        a = group_variables_by_name(_flow, ["Route"])
        self.assertEqual(flow, a["Route"])

    def test_read_vars_stringInts(self):
        plants_int = ["1", "2", "4", "5"]
        build_int = pulp.LpVariable.dicts(
            "BuildaPlant_int", plants_int, 0, 1, pulp.LpInteger
        )
        _build_int = {v.name: v for v in build_int.values()}
        b = group_variables_by_name(_build_int, ["BuildaPlant_int"])
        self.assertEqual(build_int, b["BuildaPlant_int"])

    def test_read_vars_string(self):
        plants = ["San Francisco", "Los Angeles", "Phoenix", "Denver"]
        # Creates the master problem variables of whether to build the Plants or not
        build = pulp.LpVariable.dicts("BuildaPlant", plants, 0, 1, pulp.LpInteger)
        _build = {v.name: v for v in build.values()}
        b = group_variables_by_name(
            _build, ["BuildaPlant"], replace_underscores_with_spaces=True
        )
        self.assertEqual(build, b["BuildaPlant"])

    def test_read_vars_tup_string_spaces(self):
        plants = ["San Francisco", "Los Angeles", "Phoenix", "Denver"]
        stores = ["San Diego", "Barstow", "Tucson", "Dallas"]
        plant_store = [(p, s) for p in plants for s in stores]
        flow = pulp.LpVariable.dicts("Route", plant_store, 0, None, pulp.LpInteger)
        _flow = {v.name: v for v in flow.values()}
        a = group_variables_by_name(
            _flow, ["Route"], replace_underscores_with_spaces=True
        )
        self.assertEqual(flow, a["Route"])

    def test_read_vars_twoVars(self):
        plants_int = range(1, 6)
        plants = ["San_Francisco", "Los_Angeles", "Phoenix", "Denver"]
        # Creates the master problem variables of whether to build the Plants or not
        build_int = pulp.LpVariable.dicts(
            "BuildaPlant_int", plants_int, 0, 1, pulp.LpInteger
        )
        build = pulp.LpVariable.dicts("BuildaPlant", plants, 0, 1, pulp.LpInteger)
        _build = {v.name: v for v in build_int.values()}
        _build.update({v.name: v for v in build.values()})
        b = group_variables_by_name(
            _build, ["BuildaPlant_int", "BuildaPlant"], force_number=True
        )
        self.assertEqual(build_int, b["BuildaPlant_int"])
        self.assertEqual(build, b["BuildaPlant"])
